"""World 2 dynamics simulator."""

import dataclasses
import numpy as np

from scipy.integrate import odeint

import whynot as wn
from whynot.dynamics import BaseConfig, BaseState, BaseIntervention
from whynot.simulators.world2 import tables


@dataclasses.dataclass
class Config(BaseConfig):
    # pylint: disable-msg=too-few-public-methods
    """Dynamics and simulation parameters for World2.

    Default values correspond to the standard run of World2.

    Examples
    --------
    >>> # Set the initial birth rate to 0.05
    >>> world2.config(birth_rate=0.05)

    """

    # The #: comments allow Sphinx to autodocument these parameters.
    #:
    birth_rate: float = 0.04
    #:
    death_rate: float = 0.028
    #:
    effective_capital_investment_ratio: float = 1.0
    #:
    natural_resources_usage: float = 1.0
    #:
    land_area: float = 135e6
    #:
    population_density: float = 26.5
    #:
    food_coefficient: float = 1.0
    #:
    food_normal: float = 1.0
    #:
    capital_investment_agriculture: float = 0.3
    #:
    capital_investment_generation: float = 0.05
    #:
    capital_investment_discard: float = 0.025
    #:
    pollution_standard: float = 3.6e9
    #:
    pollution: float = 1.0
    #:
    intervention_time: float = 1970
    #:
    capital_investment_in_agriculture_frac_adj_time: float = 15.0
    #:
    quality_of_life_standard: float = 1.0

    #: Time to initialize the simulation (in years).
    start_time: float = 1900
    #: Time to end the simulation (in years).
    end_time: float = 2100
    #: Time (in years) elapsed on each update of the discrete dynamics.
    delta_t: float = 0.2
    #: solver relative tolerance
    rtol: float = 1e-6
    #: solver absolute tolerance
    atol: float = 1e-6


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Class to specify interventions in World2 dynamics.

    An intervention is a subset of the configuration variables,
    and only variables passed into the constructor are modified.

    Examples
    --------
    >>> # Set the death rate to 0.06 in 1980. All other parameters unchanged.
    >>> world2.Intervention(time=1980, death_rate=0.06)

    """

    def __init__(self, time=1970, **kwargs):
        """Construct an intervention object.

        Parameters
        ----------
            time: float
                time (in years) to intervene in the simulator.
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        super(Intervention, self).__init__(Config, time, **kwargs)


@dataclasses.dataclass
class State(BaseState):
    # pylint: disable-msg=too-few-public-methods
    """World2 state. Defaults correspond to the initial state in the standard run of World2."""

    #: Total population stock
    population: float = 1.65e9
    #: Stock of natural resources
    natural_resources: float = 3.6e9 * 250
    #: Total capital investment stock
    capital_investment: float = 0.4e9
    #: Pollution stock
    pollution: float = 0.2e9
    #: Fraction of capital investment in agriculture
    capital_investment_in_agriculture: float = 0.2
    #: Initial natural resources
    initial_natural_resources: float = 3.6e9 * 250


def world2_intermediate_variables(state, config):
    """Update equations for the world2 simulaton.

    Parameters
    ----------
        state:  np.ndarray, list, or tuple
            State of the dynamics
        config: world2.Config
            Simulator configuration object that determines the coefficients

    Returns
    -------
        intermediate_variables: list
            Intermediate variables needed for world2 simulation.

    """
    (
        population,
        natural_resources,
        capital_investment,
        pollution,
        capital_investment_in_agriculture,
        initial_natural_resources,
    ) = state

    capital_investment_ratio = capital_investment / population
    crowding_ratio = population / (config.land_area * config.population_density)
    pollution_ratio = pollution / config.pollution_standard
    food_ratio = (
        tables.FOOD_POTENTIAL_FROM_CAPITAL_INVESTMENT[
            (
                capital_investment_ratio
                * capital_investment_in_agriculture
                / config.capital_investment_agriculture
            )  # capital investment ratio in agriculture
        ]  # capital investment food potential
        * tables.FOOD_FROM_CROWDING[crowding_ratio]  # crowding multiplier
        * tables.FOOD_FROM_POLLUTION[pollution_ratio]  # pollution multiplier
        * config.food_coefficient
        / config.food_normal
    )
    standard_of_living = (
        (
            capital_investment_ratio
            * tables.NATURAL_RESOURCE_EXTRACTION[
                natural_resources
                / initial_natural_resources  # fraction of natural resources remaining
            ]  # natural resources extraction multiplier
            * (1.0 - capital_investment_in_agriculture)
            / (1.0 - config.capital_investment_agriculture)
        )  # effective capital investment ratio
        / config.effective_capital_investment_ratio
    )
    death_rate_per_year = (
        config.death_rate
        * tables.DEATH_RATE_FROM_MATERIAL[standard_of_living]  # material multiplier
        * tables.DEATH_RATE_FROM_POLLUTION[pollution_ratio]  # pollution multiplier
        * tables.DEATH_RATE_FROM_FOOD[food_ratio]  # food multiplier
        * tables.DEATH_RATE_FROM_CROWDING[crowding_ratio]  # crowding multiplier
    )
    birth_rate_per_year = (
        config.birth_rate
        * tables.BIRTH_RATE_FROM_MATERIAL[standard_of_living]  # material multiplier
        * tables.BIRTH_RATE_FROM_POLLUTION[pollution_ratio]  # pollution multiplier
        * tables.BIRTH_RATE_FROM_FOOD[food_ratio]  # food multiplier
        * tables.BIRTH_RATE_FROM_CROWDING[crowding_ratio]  # crowding multiplier
    )
    natural_resources_usage_rate = (
        population
        * config.natural_resources_usage
        * tables.NATURAL_RESOURCES_FROM_MATERIAL[
            standard_of_living
        ]  # material multiplier
    )
    capital_investment_rate = (
        population
        * tables.CAPITAL_INVESTMENT_MULTIPLIER_TABLE[
            standard_of_living
        ]  # material multiplier
        * config.capital_investment_generation
    ) - capital_investment * config.capital_investment_discard
    pollution_rate = (  # pollution generation
        population
        * tables.POLLUTION_FROM_CAPITAL[capital_investment_ratio]  # capital multiplier
        * config.pollution
    ) - (  # pollution absorption
        pollution
        / tables.POLLUTION_ABSORPTION_TIME_TABLE[pollution_ratio]  # absorption time
    )

    intermediate_variables = [
        capital_investment_ratio,
        crowding_ratio,
        pollution_ratio,
        food_ratio,
        standard_of_living,
        death_rate_per_year,
        birth_rate_per_year,
        natural_resources_usage_rate,
        capital_investment_rate,
        pollution_rate,
    ]
    return intermediate_variables


def dynamics(state, time, config, intervention=None):
    """Update equations for the world2 simulaton.

    Parameters
    ----------
        state:  np.ndarray, list, or tuple
            State of the dynamics
        time:   float
        config: world2.Config
            Simulator configuration object that determines the coefficients
        intervention: world2.Intervention
            Simulator intervention object that determines when/how to update the
            dynamics.

    Returns
    -------
        ds_dt: list
            Derivative of the dynamics with respect to time

    """
    if intervention and time >= intervention.time:
        config = config.update(intervention)

    (
        population,
        natural_resources,
        capital_investment,
        pollution,
        capital_investment_in_agriculture,
        initial_natural_resources,
    ) = state

    (
        capital_investment_ratio,
        crowding_ratio,
        pollution_ratio,
        food_ratio,
        standard_of_living,
        death_rate_per_year,
        birth_rate_per_year,
        natural_resources_usage_rate,
        capital_investment_rate,
        pollution_rate,
    ) = world2_intermediate_variables(state, config)

    # Population
    delta_population = (birth_rate_per_year - death_rate_per_year) * population

    # Natural resources (negative since this is usage)
    delta_natural_resources = -natural_resources_usage_rate

    # Capital_investment
    delta_capital_investment = capital_investment_rate

    # Pollution
    delta_pollution = pollution_rate

    # Investment in agriculture
    delta_capital_investment_in_agriculture = (
        tables.CAPITAL_FRACTION_INDICATE_BY_FOOD_RATIO_TABLE[food_ratio]
        * tables.CAPITAL_INVESTMENT_FROM_QUALITY[
            (
                tables.QUALITY_OF_LIFE_FROM_MATERIAL[standard_of_living]
                / tables.QUALITY_OF_LIFE_FROM_FOOD[food_ratio]
            )  # life quality ratio
        ]  # capital investment from quality ratio
        - capital_investment_in_agriculture
    ) / config.capital_investment_in_agriculture_frac_adj_time

    ds_dt = [
        delta_population,
        delta_natural_resources,
        delta_capital_investment,
        delta_pollution,
        delta_capital_investment_in_agriculture,
        1,  # initial_natural_resources does not change
    ]
    return ds_dt


def simulate(initial_state, config, intervention=None, seed=None):
    """Run a simulation of the world2 dynamics from the given initial state.

    Parameters
    ----------
        initial_state: whynot.simulators.world2.State
            State object to initialize the simulation
        config: whynot.simulators.world2.Config
            Configuration object to determine coefficients of the dynamics.
        intervention: whynot.simulators.world2.Intervention
            (Optional) Specify what, if any, intervention to perform during execution.
        seed: int
            (Optional) The simulator is deterministic, and the seed parameter is ignored.

    Returns
    -------
        run: whynot.dynamics.Run
            Rollout sequence of states and measurement times produced by the simulator.

    """
    # Simulator is deterministic, so seed is ignored
    # pylint: disable-msg=unused-argument
    t_eval = np.arange(
        config.start_time, config.end_time + config.delta_t, config.delta_t
    )

    solution = odeint(
        dynamics,
        y0=dataclasses.astuple(initial_state),
        t=t_eval,
        args=(config, intervention),
        rtol=config.rtol,
        atol=config.atol,
    )

    states = [initial_state] + [State(*state) for state in solution[1:]]
    return wn.dynamics.Run(states=states, times=t_eval)


def quality_of_life(state, time, config, intervention=None):
    """Get the world2 quality of life metric derived from a state.

    Parameters
    ----------
        state:  whynot.simulators.world2.State or iterable representing a state.
            State of the dynamics
        time:   float
        config: world2.Config
            Simulator configuration object that determines the coefficients
        intervention: world2.Intervention
            Simulator intervention object that determines when/how to update the
            dynamics.

    Returns
    -------
        qol: float
            Quality of life metric, computed according to the world2 simulation.

    """
    if intervention and time >= intervention.time:
        config = config.update(intervention)

    if type(state) == wn.simulators.world2.State:
        state = (
            state.population,
            state.natural_resources,
            state.capital_investment,
            state.pollution,
            state.capital_investment_in_agriculture,
            state.initial_natural_resources,
        )

    (
        _,
        crowding_ratio,
        pollution_ratio,
        food_ratio,
        standard_of_living,
        _,
        _,
        _,
        _,
        _,
    ) = world2_intermediate_variables(state, config)

    qol = (
        config.quality_of_life_standard
        * tables.QUALITY_OF_LIFE_FROM_MATERIAL[
            standard_of_living
        ]  # material multiplier
        * tables.QUALITY_OF_LIFE_FROM_CROWDING[crowding_ratio]  # crowding multiplier
        * tables.QUALITY_OF_LIFE_FROM_FOOD[food_ratio]  # food multiplier
        * tables.QUALITY_OF_LIFE_FROM_POLLUTION[pollution_ratio]  # pollution multiplier
    )

    return qol



"""Experiments for the incarceration simulator."""

import numpy as np

from whynot.framework import GenericExperiment
from whynot.simulators import incarceration

__all__ = ["get_experiments", "SentenceLength"]


def get_experiments():
    """Return all available experiments for incarceration simulator.

    Returns
    -------
    list
        List of available experiments.

    """
    return [SentenceLength]


def months_in_prison(popu):
    """Return array of months in prison for each person in population."""
    return np.array([person["months_in_prison"] for person in popu.values()])


def covariates_from_population(popu):
    """Create covariates from population of agents.

    The extracted covariates are:
        * sex (1 female, 0 male)
        * age
        * number of children
        * number of siblings
        * number of friends
        * age at first birth
        * had partner (1 yes, 0 no)

    Parameters
    ----------
    popu : dict
        Dictionary specifying population of agents.

    Returns
    -------
    covariates: np.array

    """
    covariates = []
    for person in popu.values():
        covariates.append(
            [
                1 if person["sex"] == "f" else 0,
                person["death"][0] - person["birth"],
                len(person["children"]),
                len(person["siblings"]),
                len(person["friends"]),
                person["age_at_first_birth"],
                1 if person["partner"] > 0 else 0,
            ]
        )

    return np.array(covariates)


def treatments_from_population(popu):
    """Return treatment vector.

    Parameters
    ----------
    popu : dict
        Dictionary specifying population of agents.

    Returns
    -------
    treatments: np.array

    """
    treatments = []
    for person in popu.values():
        if person["harsh_sentence"]:
            treatments.append(np.mean(person["harsh_sentence"]))
        else:
            treatments.append(0)
    return np.array(treatments)


def population_to_data(popu):
    """Convert simulated population to tabular data.

    Parameters
    ----------
    popu : dict
        Dictionary specifying population of agents.

    Returns
    -------
    covariates: np.array, [num_samples, num_features]
        Covariates as numpy array.
    treatments: Binary np.array, [num_samples,]
        Treatment vector as numpy array.
    outcomes: np.array [num_samples,]
        Outcome vector as numpy arrays.

    """
    covariates = covariates_from_population(popu)
    treatments = treatments_from_population(popu)
    outcomes = months_in_prison(popu)

    return covariates, treatments, outcomes


def run_sentence_length_trial(num_samples, seed, show_progress=False, parallelize=True):
    """Run sentence experiment on incarceration simulator.

    Experiment about the effect of harsher sentences on total prison time in the
    population. Individuals are randomly assigned sentences drawn from either a
    harsher sentence distribution or a more lenient distribution. Treatment value
    in [0, 1] corresponds to the number of times an individual received a harsh
    sentence.

    Outcomes corresponds to the total amount of time an individual in the
    population spent in prison.

    The ground truth gives two vectors, one for prison times under the harsh
    policy only and one under the lenient policy only.

    Parameters
    ----------
    num_samples : int
        Unused since the simulator does not allow for varying number of agents.
    seed : int
        Random seed used for simulation
    show_progress: bool
        Show a progress bar for each run of the simulator.
    parallelize: bool
        Unused, the experiment always runs sequentially.

    Returns
    -------
    covariates: np.array, [num_samples, state_dim]
    treatment: Binary np.array, [num_samples,]
    outcomes: np.array [num_samples,]
    ground_truth_effect:

    """
    # pylint:disable-msg=unused-argument
    config = incarceration.Config(random_seed=seed, random_sentence_type=True)
    popu_random = incarceration.simulate(config, show_progress)
    covariates, treatment, outcome = population_to_data(popu_random)

    config = incarceration.Config(
        random_seed=seed, random_sentence_type=False, harsh_sentence=True
    )
    popu_harsh = incarceration.simulate(config, show_progress)

    config = incarceration.Config(
        random_seed=seed, random_sentence_type=False, harsh_sentence=False
    )
    popu_lenient = incarceration.simulate(config, show_progress)

    ground_truth = (months_in_prison(popu_harsh), months_in_prison(popu_lenient))

    return (covariates, treatment, outcome), ground_truth


# pylint: disable-msg=invalid-name
#: Randomized trial with interference between units.
SentenceLength = GenericExperiment(
    name="sentence_length_rct",
    description="Randomized controlled trial with interference between units.",
    run_method=run_sentence_length_trial,
)



"""Momoh and Fugenschuh Zika simulator initialization."""

from whynot.simulators.zika.simulator import (
    Config,
    dynamics,
    Intervention,
    simulate,
    State,
)
from whynot.simulators.zika.experiments import *
from whynot.simulators.zika.environments import *

SUPPORTS_CAUSAL_GRAPHS = True



"""Interactive environment for the credit simulator."""
import numpy as np

from whynot.gym import spaces
from whynot.gym.envs import ODEEnvBuilder, register
from whynot.simulators.credit import (
    agent_model,
    CreditData,
    Config,
    Intervention,
    simulate,
    strategic_logistic_loss,
    State,
)


def compute_reward(intervention, state, config):
    """Compute the reward based on the observed state and choosen intervention."""
    return strategic_logistic_loss(
        config, state.features, state.labels, intervention.updates["theta"],
    )


def compute_intervention(action, time):
    """Return intervention that changes the classifier parameters to action."""
    return Intervention(time=time, theta=action)


CreditEnv = ODEEnvBuilder(
    simulate_fn=simulate,
    config=Config(),
    # The initial state is the baseline features and labels in the credit dataset
    initial_state=State(features=CreditData.features, labels=CreditData.labels),
    # Action space is classifiers with the same number of parameters are
    # features.
    action_space=spaces.Box(low=-np.inf, high=np.inf, shape=(CreditData.num_features,)),
    # Observation space is the strategically adapted features and labels
    observation_space=spaces.Dict(
        {
            "features": spaces.Box(
                low=-np.inf, high=np.inf, shape=CreditData.features.shape
            ),
            "labels": spaces.Box(
                low=-np.inf, high=np.inf, shape=CreditData.labels.shape
            ),
        }
    ),
    timestep=1,
    intervention_fn=compute_intervention,
    reward_fn=compute_reward,
)

register(
    id="Credit-v0", entry_point=CreditEnv, max_episode_steps=100, reward_threshold=0,
)



"""Algorithms for causal inference."""
import whynot

try:
    import whynot_estimators

    ESTIMATORS_INSTALLED = True
except ImportError:
    ESTIMATORS_INSTALLED = False


def causal_suite(covariates, treatment, outcome, verbose=False):
    """Run a collection of causal inference algorithms on the observational dataset.

    By default, the suite only runs estimators implemented in Python
        - Ordinary least squares (ols)
        - Propensity score matching
        - Propensity weighted least squares

    Depending on the estimators installed in whynot_estimators, the suite additionally runs:
        - An IP weighting estimator (ip_weighting)
        - Matching estimators with mahalanobis distance metrics
        - Causal Forest (causal_forest)
        - TMLE          (tmle)

    Parameters
    ----------
        covariates :    `np.ndarray`
            Array of shape `num_samples` x `num_features`.
        treatment :     np.ndarray
            Boolean array of shape [num_samples] indicating treatment status for each sample.
        outcome :       np.ndarray
            Array of shape [num_sample] containing the observed outcome for each sample.
        verbose : bool
            If True, print incremental messages as estimators are executed.

    Returns
    -------
        results: dict
            Dictionary with keys denoting the name of each method and
            values the corresponding :class:`InferenceResult`:

                results['causal_forest'] -> inference_results

    """
    # Map from algorithm name to estimation function.
    methods = {
        "ols": whynot.ols,
        "propensity_score_matching": whynot.propensity_score_matching,
        "propensity_weighted_ols": whynot.propensity_weighted_ols,
    }
    if ESTIMATORS_INSTALLED:
        additional_methods = {
            "ip_weighting": whynot_estimators.ip_weighting,
            "mahalanobis_matching": whynot_estimators.matching,
            "causal_forest": whynot_estimators.causal_forest,
            "tmle": whynot_estimators.tmle,
        }
        installed = {}
        for name, estimator in additional_methods.items():
            if estimator.is_installed():
                installed[name] = estimator
        methods.update(installed)
    elif verbose:
        print("Only the base estimators are installed.")
        print(
            "To install additional additional estimators, `pip install whynot_estimators.`"
        )

    results = {}
    for name, estimator in methods.items():
        if verbose:
            print(f"Running estimator: {name}")
        try:
            inference_result = estimator.estimate_treatment_effect(
                covariates, treatment, outcome
            )
        # pylint:disable-msg=broad-except
        except Exception:
            print(f"Estimator {name} failed to run... skipping!")
            continue
        results[name] = inference_result

    return results



"""Test suite for whynot framework."""
import numpy as np
import pytest

import whynot as wn
from whynot.framework import parameter


def test_parameter():
    """Test the parameter decorator."""
    # Ccase when no standard arguments
    @parameter(name="a", default=1, values=[1, 2], description="test a")
    @parameter(name="b", default=3, values=[3, 4], description="test b")
    def foo(a, b):
        return a + b

    def check_params(listed, returned):
        assert set(listed) == set(p.name for p in returned)

    check_params(["a", "b"], wn.framework.extract_params(foo, standard_args=[]))

    @parameter(name="b", default=3, values=[3, 4], description="test b")
    def bar(a, b):
        return a + b

    # Can omit a as a parameter if a is one of the "standard args"
    check_params(["b"], wn.framework.extract_params(bar, standard_args=["a"]))

    # If there are no standard args, raise a value error since "a" is not
    # listed as a parameter
    with pytest.raises(ValueError):
        wn.framework.extract_params(bar, standard_args=[])

    # If a parameter is added, but omitted form the signature, we should raise
    with pytest.raises(ValueError):

        @parameter(name="a", default=0)
        def baz(propensity):
            return propensity

    # If a parameter is also part of standard args, we should raise
    with pytest.raises(ValueError):

        @parameter(name="rng", default=0)
        def foobar(rng):
            return 2

        wn.framework.extract_params(foobar, standard_args=["rng"])

    # If values isn't specified, we should handle things sensible
    @parameter(name="a", default=1)
    @parameter(name="b", default=3, values=[3, 4])
    @parameter(name="c", default=4, values=np.arange(0.05, 0.95, 0.1))
    def values_test(a, b, c):
        return a + b + c

    param_collection = wn.framework.extract_params(values_test, standard_args=[])

    sampled = param_collection.sample()
    assert sampled["a"] == 1
    print(sampled)


def test_parameter_collection():
    """Test the parameter collection object."""
    from whynot.framework import ExperimentParameter

    p1 = ExperimentParameter(name="p1", default=1)
    collection = wn.framework.ParameterCollection(params=[p1])
    collection.add_parameter(ExperimentParameter(name="p2", default=2))

    projected = collection.project({"p1": 10})
    assert projected["p1"] == 10
    assert projected["p2"] == 2

    p3 = ExperimentParameter(name="p3", default=3)
    p4 = ExperimentParameter(name="p4", default=4)
    collection2 = wn.framework.ParameterCollection(params=[p3, p4])

    collection += collection2

    assert collection["p3"].default == 3

    # Ensure names are unique
    a = ExperimentParameter(name="a", default=3)
    aa = ExperimentParameter(name="a", default=4)
    with pytest.raises(ValueError):
        wn.framework.ParameterCollection(params=[a, aa])

    # Ensure names are unique after adding parameter
    with pytest.raises(ValueError):
        coll = wn.framework.ParameterCollection(params=[a])
        coll.add_parameter(aa)

    # Ensure names are unique if you add two collections
    with pytest.raises(ValueError):
        coll_one = wn.framework.ParameterCollection(params=[a])
        coll_two = wn.framework.ParameterCollection(params=[aa])
        new = coll_one + coll_two



"""Experiments for Zika simulator."""

from whynot.dynamics import DynamicsExperiment
from whynot.framework import parameter
from whynot.simulators import zika


__all__ = ["get_experiments", "ZikaRCT"]


def get_experiments():
    """Return all experiments for zika."""
    return [ZikaRCT]


def sample_initial_states(rng):
    """Sample initial state by randomly perturbing the default state."""
    state = zika.State().values()
    state *= rng.uniform(low=0.95, high=1.05, size=state.shape)
    return zika.State(*state)


def mixed_treatment_intervention():
    """Apply mixed strategy on all controls starting at time 0."""
    return zika.Intervention(
        time=0,
        treated_bednet_use=0.5,
        condom_use=0.5,
        treatment_of_infected=0.5,
        indoor_spray_use=0.5,
    )


##################
# RCT Experiment
##################
# pylint: disable-msg=invalid-name
#: Experiment on effect of mixed treatment policy on infections in 20 days
ZikaRCT = DynamicsExperiment(
    name="ZikaRCT",
    description="Study effect of mixed treatment policy on infections in 20 days.",
    simulator=zika,
    simulator_config=zika.Config(start_time=0, end_time=20),
    intervention=mixed_treatment_intervention(),
    state_sampler=sample_initial_states,
    propensity_scorer=0.5,
    outcome_extractor=lambda run: run[19].symptomatic_humans,
    covariate_builder=lambda run: run.initial_state.values(),
)



"""Benchmark execution time of all of the simulators."""
import time

import numpy as np

import whynot as wn


def benchmark_dynamic_simulator(simulator, name, num_trials=10):
    """Run dynamical system simulator repeatedly and report execution time statistics."""
    timings = []
    for _ in range(num_trials):
        initial_state = simulator.State()
        config = simulator.Config()
        start_time = time.perf_counter()
        simulator.simulate(initial_state, config)
        timings.append(time.perf_counter() - start_time)
    print(
        f"{name.upper()}:\t average={np.mean(timings):.3f} s, std={np.std(timings):.3f} s, "
        f"max={np.max(timings):.3f} s, (trials={num_trials})"
    )


def benchmark_agent_based_model(simulator, name, num_agents=100, num_trials=10):
    """Run the agent-based model repeatedly and report execution time statistics."""
    timings = []
    for _ in range(num_trials):
        agents = [simulator.Agent() for _ in range(num_agents)]
        config = simulator.Config()
        start_time = time.perf_counter()
        simulator.simulate(agents, config)
        timings.append(time.perf_counter() - start_time)
    print(
        f"{name.upper()}:\t average={np.mean(timings):.3f} s, std={np.std(timings):.3f} s, "
        f"max={np.max(timings):.3f} s, (trials={num_trials})"
    )


def benchmark_all():
    """Run benchmarks for each simulator."""
    benchmark_dynamic_simulator(wn.dice, "dice")
    benchmark_dynamic_simulator(wn.hiv, "hiv")
    benchmark_dynamic_simulator(wn.lotka_volterra, "lotka-volterra")
    benchmark_dynamic_simulator(wn.opioid, "opioid")
    benchmark_dynamic_simulator(wn.world2, "world2")
    benchmark_dynamic_simulator(wn.world3, "world3")

    benchmark_agent_based_model(wn.civil_violence, "civil_violence")


if __name__ == "__main__":
    benchmark_all()



"""Lotka-Volterra initialization."""

from whynot.simulators.lotka_volterra.simulator import Config, Intervention, dynamics
from whynot.simulators.lotka_volterra.simulator import simulate, State
from whynot.simulators.lotka_volterra.experiments import *

SUPPORTS_CAUSAL_GRAPHS = True



"""Reinforcement learning environments for the opioid epidemic simulator."""
import numpy as np

from whynot.gym import spaces
from whynot.gym.envs import ODEEnvBuilder, register
from whynot.simulators.opioid import Config, Intervention, simulate, State


def get_intervention(action, time):
    """Return the intervention in the simulator required to take action."""
    action_to_intervention_map = {
        0: Intervention(time=time, nonmedical_incidence=0.0, illicit_incidence=0.0),
        1: Intervention(time=time, nonmedical_incidence=0.0, illicit_incidence=0.05),
        2: Intervention(time=time, nonmedical_incidence=0.05, illicit_incidence=0.0),
        3: Intervention(time=time, nonmedical_incidence=0.05, illicit_incidence=0.05),
    }
    return action_to_intervention_map[action]


def get_reward(intervention, state, config, time):
    """Compute the reward based on the observed state and choosen intervention."""
    # pylint: disable-msg=invalid-name
    # Penalty for overdose death.
    C = 1000
    # Costs for reducing nonmedical users and illicit users per user.
    C1 = 1
    C2 = 5
    deaths = state.nonmedical_users * config.nonmedical_overdose
    deaths += state.oud_users * config.oud_overdose[time]
    deaths += state.illicit_users * config.illicit_overdose[time]
    reward = -C * deaths
    reward -= C1 * intervention.updates["nonmedical_incidence"] * state.nonmedical_users
    reward -= C2 * intervention.updates["illicit_incidence"] * state.illicit_users
    return reward


def observation_space():
    """Return the observation space. The state is (nonmedical_users, oud_useres, illicit_users)."""
    state_dim = State.num_variables()
    state_space_low = np.zeros(state_dim)
    state_space_high = np.inf * np.ones(state_dim)
    return spaces.Box(state_space_low, state_space_high)


OpioidEnv = ODEEnvBuilder(
    simulate_fn=simulate,
    config=Config(),
    initial_state=State(),
    # In this environment we define 4 actions:
    #   - Do nothing
    #   - Reduce nonmedical opioid use by 5%
    #   - Reduce illicit opioid use by 5%
    #   - Reduce both by 5%
    action_space=spaces.Discrete(4),
    observation_space=observation_space(),
    timestep=1.0,
    intervention_fn=get_intervention,
    reward_fn=get_reward,
)

register(
    id="opioid-v0",
    entry_point=OpioidEnv,
    # The simulator starts in 2002 and ends in 2030.
    max_episode_steps=28,
    reward_threshold=0,
)



"""Unit tests for Dice simulator."""
import os

import numpy as np
import pandas as pd

import whynot as wn


def test_gams():
    """Ensure calling default simulator has results reflective of the GAMS implementaiton"""
    # Load results from standard GAMS results
    gams_file = os.path.realpath(__file__).rsplit("/", 1)[0] + "/gamsBaseline.csv"
    gams = pd.read_csv(gams_file).replace([np.inf, -np.inf], np.nan).fillna(0)
    gams.columns = ["Variable", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
    gams.drop("Variable", axis=1, inplace=True)

    # Compute simulator results
    config = wn.dice.Config(numPeriods=10, ifopt=0)
    initial_state = wn.dice.State()
    run = wn.dice.simulate(initial_state, config, stochastic=False)

    # Run without randomness over the config parameters.
    sim_dict = {
        var: [getattr(state, var) for state in run.states[1:]]
        for var in initial_state.variables
    }
    sim = (
        pd.DataFrame.from_dict(sim_dict, orient="index", columns=gams.columns)
        .replace([np.inf, -np.inf], np.nan)
        .fillna(0)
    )

    np.testing.assert_allclose(sim, gams, rtol=1e-4, atol=5e-2, equal_nan=False)


if __name__ == "__main__":
    test_gams()



"""Simulate and analyze runs from the civil violence model."""
import dataclasses

from mesa.datacollection import DataCollector
import numpy as np

from whynot.simulators.civil_violence.model import CivilViolenceModel


@dataclasses.dataclass
class Agent:
    # pylint: disable-msg=too-few-public-methods
    """Covariates for a single agent in the simulation.

    Examples
    --------
    >>> # An agent with low risk aversion and high hardship
    >>> civil_violence.Agent(hardship=0.99, risk_aversion=0.01)

    """

    #: How aggrieved is the agent by external circumstances.
    hardship: float = 0.5
    #: How legitimate does the agent perceive the current regime.
    legitimacy: float = 0.5
    #: Threshold above which agent starts to openly rebell
    active_threshold: float = 0.1
    #: How likely the agent is to rebel for a fixed greviance level.
    risk_aversion: float = 0.5
    #: How many adjacent squares an agent sees and uses to determine probability of arrest.
    vision: int = 3


@dataclasses.dataclass
class Config:
    # pylint: disable-msg=too-few-public-methods
    """Metaparameters for a single run of the civil violence simulator."""

    #: Vertical grid size
    grid_height: int = 50
    #: Horizontal grid size
    grid_width: int = 50
    #: What fraction of the agents are police
    cop_fraction: float = 0.05
    #: How many adjacent squares a cop can see when determining who to arrest.
    cop_vision: int = 5

    # What's the longest time a citizen can stay in jail?
    max_jail_term: int = 5
    #: How strongly other agent grievances affect each agent in prison.
    prison_interaction: float = 0.1
    #: A fixed parameter to calibrate arrest likelihood.
    arrest_prob_constant: float = 2.3


def count_type_citizens(model, condition, exclude_jailed=True):
    """Count agents as either Quiescent/Active."""
    count = 0
    for agent in model.schedule.agents:
        if agent.breed == "cop":
            continue
        if exclude_jailed and agent.jail_sentence:
            continue
        if agent.condition == condition:
            count += 1
    return count


def count_jailed(model):
    """Count number of jailed agents."""
    count = 0
    for agent in model.schedule.agents:
        if agent.breed == "citizen" and agent.jail_sentence:
            count += 1
    return count


def simulate(agents, config, seed=None, max_steps=1000):
    """Simulate a run of the civil violence model.

    Parameters
    ----------
        agents: list
            List of whynot.simulators.civil_violence.Agent to populate the model
        config: whynot.simulators.civil_violence.Config
            Simulation parameters
        seed: int
            (Optional) Seed for all randomness in model setup and execution.
        max_steps: int
            Maximum number of steps to run the civil_violence model.

    Returns
    -------
        observations: pd.DataFrame
            Pandas dataframe containing the "observations" recorded for each
            agent. Observations are defined in the `agent_reporter` and include
            agent attributes along with:
                "pos" # position on the grid
                "jail_sentence" # agent's jail sentence at model end
                "condition"  # agent's condition (rebelling or acquiesent) at model end
                "arrest_probability" # agent's probability of arrest
                "arrests" # number of time agent has been arrested
                "days_active"  # how long as the agent spent in rebellion

    """
    # Ensure everything will fit on the grid
    num_cells = config.grid_height * config.grid_width
    num_cops = int(np.floor(len(agents) * config.cop_fraction))

    assert len(agents) + num_cops < num_cells

    model = CivilViolenceModel(
        height=config.grid_height,
        width=config.grid_width,
        cop_vision=config.cop_vision,
        max_jail_term=config.max_jail_term,
        prison_interaction=config.prison_interaction,
        arrest_prob_constant=config.arrest_prob_constant,
        max_steps=max_steps,
        seed=seed,
    )
    # Place agents on grid
    for i, agent in enumerate(agents):
        model.add_agent(
            i,
            model.find_empty(),
            agent.hardship,
            agent.legitimacy,
            agent.risk_aversion,
            agent.active_threshold,
            agent.vision,
        )

    for i in range(num_cops):
        model.add_cop(i + len(agents), model.find_empty())

    # Which attributes to report
    agent_reporters = {
        "pos": "pos",
        "breed": "breed",
        "jail_sentence": "jail_sentence",
        "condition": "condition",
        "arrest_probability": "arrest_probability",
        "arrests": "arrests",
        "hardship": "hardship",
        "regime_legitimacy": "regime_legitimacy",
        "days_active": "days_active",
        "risk_aversion": "risk_aversion",
        "threshold": "threshold",
        "arrest_parameter": "arrest_parameter",
        "vision": "vision",
    }

    datacollector = DataCollector(agent_reporters=agent_reporters)
    while model.running:
        model.step()
    datacollector.collect(model)
    dataframe = datacollector.get_agent_vars_dataframe()
    observations = dataframe[dataframe.breed == "citizen"].drop(columns="breed")
    return observations



"""Experiments on the Lotka-Volterra model."""
import numpy as np

from whynot.dynamics import DynamicsExperiment
from whynot.framework import parameter
from whynot.simulators import lotka_volterra


__all__ = ["get_experiments", "RCT", "Confounding", "UnobservedConfounding"]


def get_experiments():
    """Return all of the Lotka-Volterra experiments."""
    return [RCT, Confounding, UnobservedConfounding]


def sample_initial_states(rng):
    """Sample an initial state for the LV model."""
    rabbits = rng.randint(10, 100)
    # Ensure the number of rabbits is greater than number of foxes.
    foxes = rng.uniform(0.1, 0.8) * rabbits
    return lotka_volterra.State(rabbits=rabbits, foxes=foxes)


@parameter(
    name="window",
    default=10,
    values=[5, 10, 20],
    description="Number of years to inspect outcome.",
)
@parameter(
    name="outcome_year",
    default=80,
    values=[30, 50, 80],
    description="Year to measure outcome.",
)
def lv_outcome_extractor(run, window, outcome_year):
    """Return the minimum fox population over the window before the outcome year."""
    return np.min(
        [run[year].foxes for year in range(outcome_year - window, outcome_year)]
    )


###########################
# RCT Experiment
###########################
@parameter(
    name="propensity",
    default=0.1,
    values=np.linspace(0.05, 0.5, 10),
    description="Probability of treatment",
)
def rct_propensity(propensity):
    """Return constant propensity for RCT."""
    return propensity


#: RCT experiment to determine effect of reducing number of caught rabbits that create a fox.
RCT = DynamicsExperiment(
    name="lotka_volterra_rct",
    description="A RCT to determine effect of reducing rabbits needed to sustain a fox.",
    simulator=lotka_volterra,
    simulator_config=lotka_volterra.Config(fox_growth=0.75),
    intervention=lotka_volterra.Intervention(time=30, fox_growth=0.4),
    state_sampler=sample_initial_states,
    propensity_scorer=rct_propensity,
    outcome_extractor=lv_outcome_extractor,
    covariate_builder=lambda run: run.initial_state.values(),
)


##########################
# Confounding Experiments
##########################
@parameter(
    name="propensity",
    default=0.9,
    values=np.linspace(0.5, 0.99, 10),
    description="Probability of treatment for group with low fox population.",
)
def confounded_propensity_scores(untreated_run, propensity=0.9):
    """Return confounded treatment assignment probability.

    Treatment increases fox population growth. Therefore, we're assume
    treatment is more likely for runs with low initial fox population.
    """
    if untreated_run.initial_state.foxes < 20:
        return propensity
    return 1.0 - propensity


# pylint: disable-msg=invalid-name
#: Observational experiment with confounding due to initial fox population.
Confounding = DynamicsExperiment(
    name="lotka_volterra_confounding",
    description=(
        "Determine effect of reducing rabbits needed to sustain a fox."
        "Treament confounded by initial fox population."
    ),
    simulator=lotka_volterra,
    simulator_config=lotka_volterra.Config(fox_growth=0.75),
    intervention=lotka_volterra.Intervention(time=30, fox_growth=0.4),
    state_sampler=sample_initial_states,
    propensity_scorer=confounded_propensity_scores,
    outcome_extractor=lv_outcome_extractor,
    covariate_builder=lambda run: run.initial_state.values(),
)


# pylint: disable-msg=invalid-name
#: Observational experiment with unobserved confounding due to initial fox population.
UnobservedConfounding = DynamicsExperiment(
    name="lotka_volterra_unobserved_confounding",
    description=(
        "Determine effect of reducing rabbits needed to sustain a fox."
        "Treament confounded by initial fox population, which is not observed."
    ),
    simulator=lotka_volterra,
    simulator_config=lotka_volterra.Config(fox_growth=0.75),
    intervention=lotka_volterra.Intervention(time=30, fox_growth=0.4),
    state_sampler=sample_initial_states,
    propensity_scorer=confounded_propensity_scores,
    outcome_extractor=lv_outcome_extractor,
    covariate_builder=lambda run: np.array([run.initial_state.rabbits]),
)



"""Experiments for world2 simulator."""
import numpy as np

from whynot.dynamics import DynamicsExperiment
from whynot.framework import parameter
from whynot.simulators import world2

__all__ = ["get_experiments", "RCT", "BiasedTreatment", "Mediation"]


def get_experiments():
    """Return all experiments for world2."""
    return [RCT, BiasedTreatment, Mediation]


def sample_initial_states(rng):
    """Sample an initial world2 state by perturbing the default."""
    state = world2.State()
    state.population *= rng.uniform(0.75, 2)
    state.natural_resources *= rng.uniform(0.25, 10.0)
    state.capital_investment *= rng.uniform(0.5, 2.0)
    state.pollution *= rng.uniform(0.5, 2)
    state.capital_investment_in_agriculture *= rng.uniform(0.5, 1.5)
    state.initial_natural_resources = state.natural_resources
    return state


###########################
# RCT Experiment
###########################
#: RCT experiment for world2.
RCT = DynamicsExperiment(
    name="world2_rct",
    description="RCT to measure the effect of increasing capital investment in 1970.",
    simulator=world2,
    simulator_config=world2.Config(),
    intervention=world2.Intervention(time=1970, capital_investment_generation=0.06),
    state_sampler=sample_initial_states,
    propensity_scorer=0.25,
    outcome_extractor=lambda run: run[2000].population,
    covariate_builder=lambda run: run.initial_state.values(),
)


###########################
# Treatment Bias Experiment
###########################
@parameter(
    name="treatment_bias",
    default=0.2,
    values=[0.1, 0.2, 0.5, 0.9],
    description="How much to bias treatment assignment",
)
@parameter(
    name="treatment_percentage",
    default=0.2,
    values=[0.05, 0.2, 0.5, 0.9],
    description="What fraction of the units should be treated.",
)
def biased_treatment_propensity(untreated_runs, treatment_bias, treatment_percentage):
    """Compute propensity scores as a function of the bias."""
    # Sort runs by initial quantity of natural resources
    sorted_idxs = sorted(
        range(len(untreated_runs)),
        key=lambda idx: untreated_runs[idx].initial_state.natural_resources,
        reverse=True,
    )

    # More likely to treat units with top initial quantity of natural resources
    top_resource_idxs = sorted_idxs[int(treatment_percentage * len(sorted_idxs)) :]

    propensities = (1.0 - treatment_bias) * np.ones(len(untreated_runs))
    propensities[top_resource_idxs] = treatment_bias

    return propensities


# pylint: disable-msg=invalid-name
#: Observational experiment with confounding for world2.
BiasedTreatment = DynamicsExperiment(
    name="world2_biased_treatment",
    description="Experiment with treatment bias on World 2.",
    simulator=world2,
    simulator_config=world2.Config(),
    intervention=world2.Intervention(time=1970, capital_investment_generation=0.06),
    state_sampler=sample_initial_states,
    propensity_scorer=biased_treatment_propensity,
    outcome_extractor=lambda run: run[2000].population,
    covariate_builder=lambda run: run.initial_state.values(),
)


######################
# Mediation Experiment
######################
@parameter(
    name="birth_rate_intervention",
    default=0.02,
    values=[0.039, 0.02, 0.01],
    description="Decrease in birth rate in the intervention year.",
)
def mediation_intervention(birth_rate_intervention):
    """Return the treatment config for the mediation experiment.

    Provides an example showing how the framework supports parameterized
    configurations.
    """
    base_birth_rate = world2.Config().birth_rate
    assert np.less_equal(birth_rate_intervention, base_birth_rate)
    return world2.Intervention(time=1970, birth_rate=birth_rate_intervention)


def mediation_propensity_scores(intervention, untreated_runs):
    """Probability of treating each unit.

    Units with the largest populations are more likely to be treated.
    """
    populations = [run[intervention.time].population for run in untreated_runs]
    upper_quantile = np.quantile(populations, 0.9)
    propensities = 0.05 * np.ones(len(untreated_runs))
    propensities[populations > upper_quantile] = 0.9
    return propensities


@parameter(
    name="mediation_year",
    default=1980,
    values=[1980, 2000, 2030, 2030],
    description="Year to choose the mediation states.",
)
@parameter(
    name="num_mediators",
    default=3,
    values=[0, 1, 2, 3, 4],
    description="How many mediating states to choose.",
)
def mediation_covariates(run, intervention, mediation_year, num_mediators):
    """Return covariates including mediators."""
    confounders = run[intervention.time].values()
    mediators = run[mediation_year].values()
    # Use only a subset of the mediators to avoid blocking all paths
    return np.concatenate([confounders, mediators[:num_mediators]])


def mediation_outcome_extractor(run, config, intervention):
    return world2.quality_of_life(
        state=run[2030], time=2030, config=config, intervention=intervention
    )


# pylint: disable-msg=invalid-name
#: Observational experiment with mediation for world2.
Mediation = DynamicsExperiment(
    name="world2_mediation",
    description="Create mediation by adding covariates after intervention.",
    simulator=world2,
    simulator_config=world2.Config(),
    intervention=mediation_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=mediation_propensity_scores,
    outcome_extractor=mediation_outcome_extractor,
    covariate_builder=mediation_covariates,
)

if __name__ == "__main__":
    RCT.run(num_samples=10)



"""Utilities for loading and manipulating FICO data.

Borrowed from:
    https://github.com/ecreager/delayedimpact-scm-shareable
"""
import os
from typing import Callable, Iterable

import numpy as np
import pandas as pd

Index = pd.Index
Array = np.ndarray
DataFrame = pd.DataFrame


#############################
# FICO Data loading
#############################

PERF = "transrisk_performance_by_race_ssa.csv"
CDF_BY_RACE = "transrisk_cdf_by_race_ssa.csv"
OVERALL = "totals.csv"

FILES = dict(cdf_by_race=CDF_BY_RACE, performance_by_race=PERF, overview=OVERALL)


def cleanup_frame(frame):
    """Make the columns have better names, and ordered in a better order."""
    frame = frame.rename(columns={"Non- Hispanic white": "White"})
    frame = frame.reindex(columns=["Asian", "White", "Hispanic", "Black"])
    return frame


def read_totals(data_dir):
    """Read the total number of people of each race."""
    frame = cleanup_frame(
        pd.read_csv(os.path.join(data_dir, FILES["overview"]), index_col=0)
    )
    return {r: frame[r]["SSA"] for r in frame.columns}


def convert_percentiles(idx):
    """Convert percentiles."""
    pdf = [
        (300, 2.1),
        (350, 4.2),
        (400, 5.4),
        (450, 6.5),
        (500, 7.9),
        (550, 9.6),
        (600, 12.0),
        (650, 13.8),
        (700, 17.0),
        (750, 15.8),
        (800, 5.7),
        (850, 0),
    ]

    def convert_one(x):
        partial = 0
        for ((v, s), (v2, _)) in zip(pdf, pdf[1:]):
            if partial + s >= x:
                return v + (v2 - v) * (x - partial) / s
            partial += s

    return np.array(list(map(convert_one, idx)))


def parse_data(data_dir, filenames=None):
    """Parse sqf data set."""
    if filenames is None:
        filenames = [FILES["cdf_by_race"], FILES["performance_by_race"]]

    cdfs = cleanup_frame(pd.read_csv(os.path.join(data_dir, filenames[0]), index_col=0))
    performance = 100 - cleanup_frame(
        pd.read_csv(os.path.join(data_dir, filenames[1]), index_col=0)
    )
    return (cdfs / 100.0, performance / 100.0)


def get_FICO_data(data_dir, do_convert_percentiles=True):
    """Get FICO data in desired format."""
    data_pair = parse_data(data_dir)
    totals = read_totals(data_dir)

    if do_convert_percentiles:
        for v in data_pair:
            v.index = convert_percentiles(v.index)
    cdfs = data_pair[0]
    performance = data_pair[1]
    return cdfs, performance, totals


########################
# Manipulating FICO data
########################


def _find_nearest_indices(idx: Index, targets: Array) -> Array:
    """Convert 1d array into array of nearest index values."""
    idx = np.asarray(idx)
    idx_ = np.expand_dims(idx, -1)
    targets = np.expand_dims(targets, 0)
    i = (np.abs(idx_ - targets)).argmin(axis=0)
    return idx[i]


def _get_pmf(cdf):
    """Convert CDF into PMF."""
    pis = np.zeros(cdf.size)
    pis[0] = cdf[0]
    for score in range(cdf.size - 1):
        pis[score + 1] = cdf[score + 1] - cdf[score]
    return pis


def _loan_repaid_probs_factory(
    repay_df: DataFrame, scores: Index
) -> Iterable[Callable[[Array], Array]]:
    """Given performance pd.DataFrame, construct mapping from X to p(Y|X)."""

    def repaid_probs_fn(query_scores: Array) -> Array:
        if isinstance(query_scores, np.ndarray):
            nearest_scores = _find_nearest_indices(scores, query_scores)
            return repay_df[nearest_scores].values

        query_score = query_scores
        nearest_score = scores[scores.get_loc(query_score, method="nearest")]
        return repay_df[nearest_score]

    return repaid_probs_fn


def get_data_args(data_dir="data"):
    """Return objects that specify p(A), p(X|A), p(Y|X,A)."""

    all_cdfs, performance, totals = get_FICO_data(data_dir)
    # NOTE: we drop last column to make the CDFs invertible ####################
    all_cdfs = all_cdfs.drop(all_cdfs.index[-1])
    performance = performance.drop(performance.index[-1])
    ############################################################################
    cdfs = all_cdfs[["White", "Black"]]

    cdf_B = cdfs["White"].values
    cdf_A = cdfs["Black"].values

    repay_B = performance["White"]
    repay_A = performance["Black"]

    scores = cdfs.index
    scores_list = scores.tolist()

    # to populate group distributions
    # to get (batchable) loan repay probabilities for a given score
    loan_repaid_probs = [
        _loan_repaid_probs_factory(repay_A, scores),
        _loan_repaid_probs_factory(repay_B, scores),
    ]

    # to get score at a given probabilities
    inv_cdfs = get_inv_cdf_fns(cdfs)

    # get probability mass functions of each group
    pi_A = _get_pmf(cdf_A)
    pi_B = _get_pmf(cdf_B)
    pis = np.vstack([pi_A, pi_B])

    # demographic statistics
    group_ratio = np.array((totals["Black"], totals["White"]))
    group_size_ratio = group_ratio / group_ratio.sum()

    rate_indices = (list(reversed(1 - cdf_A)), list(reversed(1 - cdf_B)))

    return inv_cdfs, loan_repaid_probs, pis, group_size_ratio, scores_list, rate_indices


def get_inv_cdf_fns(cdfs: DataFrame) -> Iterable[Callable[[Array], Array]]:
    """Convert DataFrame of cdfs into list of (batched) inv. cdf lambda fns."""

    def inv_cdf_factory(cdfs_df: DataFrame, key: str) -> Callable[[Array], Array]:
        """Given cdfs pd.DataFrame & key=A, make mapping from P(Y|X,A) to X."""
        series = pd.Series(cdfs_df[key].index.values, index=cdfs[key].values)
        index = series.index

        def repaid_probs_fn(query_probs: Array) -> Array:
            if isinstance(query_probs, np.ndarray):
                nearest_scores = _find_nearest_indices(index, query_probs)
                return series[nearest_scores].values

            query_prob = query_probs
            nearest_prob = index[index.get_loc(query_prob, method="nearest")]
            return series[nearest_prob]

        return repaid_probs_fn

    inv_cdfs = [inv_cdf_factory(cdfs, "Black"), inv_cdf_factory(cdfs, "White")]

    return inv_cdfs


def get_marginal_loan_repaid_probs(data_dir="data"):
    """Return object that specifies distn p(Y|X)."""

    all_cdfs, performance, totals = get_FICO_data(data_dir)
    # NOTE: we drop last column to make the CDFs invertible ####################
    all_cdfs = all_cdfs.drop(all_cdfs.index[-1])
    performance = performance.drop(performance.index[-1])
    ############################################################################
    cdfs = all_cdfs[["White", "Black"]]

    repay_B = performance["White"]
    repay_A = performance["Black"]

    scores = cdfs.index

    # demographic statistics
    group_ratio = np.array((totals["Black"], totals["White"]))
    group_size_ratio = group_ratio / group_ratio.sum()
    p_A, p_B = group_size_ratio

    repay_marginal = repay_A * p_A + repay_B * p_B
    loan_repaid_probs = _loan_repaid_probs_factory(repay_marginal, scores)
    loan_repaid_probs = [
        loan_repaid_probs,
        loan_repaid_probs,
    ]  # return a copy for each group for compatability with other utils

    return loan_repaid_probs



"""Implementation of Momoh and Fügenschuh ODE simulator for Zika prevention.

Momoh, Abdulfatai A., and Armin Fügenschuh. "Optimal control of intervention
strategies and cost effectiveness analysis for a Zika virus model." Operations
Research for Health Care 18 (2018): 99-111.

https://www.sciencedirect.com/science/article/pii/S2211692316301084#!
"""

import dataclasses
import numpy as np

from scipy.integrate import odeint

import whynot as wn
from whynot.dynamics import BaseConfig, BaseState, BaseIntervention


@dataclasses.dataclass
class Config(BaseConfig):
    # pylint: disable-msg=too-few-public-methods
    """Parameters for the simulation dynamics."""

    # Simulation parameters
    #: Recruitment rate of humans into susceptible population
    lambda_h: float = 0.000011
    #: Natural human death rate
    mu_h: float = 1.0 / (360 * 60)
    #: Recovered humans loss of immunity
    varphi_h: float = 0.02
    #: Spontaneous individual recovery
    phi: float = 0.05
    #: Rate of recovery with temporary immunity
    nu: float = 0.023
    #: Breakthrough rate of humans from exposed to infected
    alpha_h: float = 0.2
    #: Disease induced death rate
    delta_h: float = 0.003
    #: Relative human-human contact rate of asymptomatic infected
    c: float = 0.05
    #: Relative human-human contact rate of symptomatic infected
    kappa: float = 0.05
    #: Prob of transmission per contact by asymptomatic infected through sexual activity
    beta_a: float = 0.6
    #: Prob of transmission per contact by symptomatic infected through sexual activity
    beta_s: float = 0.3
    #: Prob of transmission per contact by an infectious mosquito
    beta_1: float = 0.4
    #: Prob of transmission per contact by an infected human
    beta_2: float = 0.5
    #: Per capital biting rate of mosquitos
    epsilon: float = 0.5
    #: Contact rate of mosquito per human per unit time
    rho: float = 0.1
    #: Recruitment rate of mosquitoes
    lambda_v: float = 0.071
    #: Natural death rate of mosquitoes
    mu_v: float = 1.0 / 14.0
    #: Breakthrough rate of mosquitoes from exposed to infectious
    alpha_v: float = 0.1
    #: Rate constant due to indoor residual spray
    theta: float = 0.75
    #: Rate constant due to treatment effort
    tau: float = 0.15

    # Treatment parameters

    # Treated bednets and condom use represent the faction of susceptible and
    # asymptomatic individuals who use them to minimize mosquito to human or
    # human to human contact with the virus. IRS spray use affects the whole
    # mosquito population by increasing its mortality rate.
    # All variables are [0, 1].

    #: Control measure through use of treated bednets
    treated_bednet_use: float = 0.0
    #: Control measure through use of condoms
    condom_use: float = 0.0
    #: Control measure through treatment of infected humans
    treatment_of_infected: float = 0.0
    #: Control measure through use of indoor residual spray
    indoor_spray_use: float = 0.0

    #: Simulation start time (in day)
    start_time: float = 0
    #: Simulation end time (in days)
    end_time: float = 100
    #: How frequently to measure simulator state
    delta_t: float = 1.0
    #: solver relative tolerance
    rtol: float = 1e-6
    #: solver absolute tolerance
    atol: float = 1e-6

    @property
    def recovery_rate(self):
        """Rate of recovery from Zika."""
        # spontaneous recovery + recovery from treatment
        return self.phi + self.tau * self.treatment_of_infected

    @property
    def mosquito_human_infection_rate(self):
        """Infection rate of humans from mosquito bites."""
        return (
            self.beta_1  # probability of disease transmission from mosquitoes
            * self.epsilon  # mosquito biting rate
            * self.rho  # human/mosquito contact rate
            * (1.0 - self.treated_bednet_use)
        )

    @property
    def human_mosquito_infection_rate(self):
        """Infection rate of mosquitos from humans."""
        return (
            self.beta_2  # probability of disease transmission to mosquitoes
            * self.epsilon  # mosquito biting rate
            * self.rho  # human/mosquito contact rate
            * (1.0 - self.treated_bednet_use)
        )

    @property
    def asymptomatic_infection_rate(self):
        """Infection from sexual contact with asymptomatic infected humans."""
        return (
            self.beta_a  # prob of transmission from sexual contact with asymptomatic
            * self.c  # relative human/human contact rate of asymptomatic infected
            * (1.0 - self.condom_use)
        )

    @property
    def symptomatic_infection_rate(self):
        """Infection from sexual contact with symptomatic infected humans."""
        return (
            self.beta_s  # prob of transmission from sexual contact with symptomatic
            * self.kappa  # relative human/human contact rate of symptomatic infected
            * (1.0 - self.condom_use)
        )


@dataclasses.dataclass
class State(BaseState):
    # pylint: disable-msg=too-few-public-methods
    """State of the Zika simulator.

    The default state corresponds to a small community with
    a Zika outbreak. Most of the inhabitants are susceptible, a small number
    are exposed, but asymptomatic, and a tiny fraction ~1% are infected and
    symptomatic.
    """

    #: Number of susceptible humans
    susceptible_humans: float = 750.0
    #: Number of asymptomatic infected human
    asymptomatic_humans: float = 250.0
    #: Number of symptomatic infected humans
    symptomatic_humans: float = 10.0
    #: Number of recovered humans
    recovered_humans: float = 20.0
    #: Number of susceptible mosquitoes
    susceptible_mosquitos: float = 10000.0
    #: Number of exposed mosquitoes
    exposed_mosquitos: float = 500.0
    #: Number of infectious mosquitoes
    infectious_mosquites: float = 100.0
    #: Total number of humans
    human_population: float = 1030.0
    #: Total number of mosquitoes
    mosquito_population: float = 10600


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention in the Zika model.

    Examples
    --------
    >>> # Starting in step 10, set bed net use to 0.5 (leaving other variables unchanged)
    >>> Intervention(time=10, treated_bednet_use=0.5)

    """

    def __init__(self, time, **kwargs):
        """Specify an intervention in the dynamical system.

        Parameters
        ----------
            time: int
                Time of the intervention (days)
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        super(Intervention, self).__init__(Config, time, **kwargs)


def dynamics(state, time, config, intervention=None):
    """Update equations for the Zika simulaton.

    Parameters
    ----------
        state:  np.ndarray, list, or tuple
            State of the dynamics
        time:   float
        config: hiv.Config
            Simulator configuration object that determines the coefficients
        intervantion: hiv.Intervention
            Simulator intervention object that determines when/how to update the
            dynamics.

    Returns
    -------
        ds_dt: list
            Derivative of the dynamics with respect to time

    """
    if intervention and time >= intervention.time:
        config = config.update(intervention)

    # Keep notation roughly consistent with the original paper
    # pylint: disable-msg=invalid-name
    (
        S_h,  # susceptible humans
        A_h,  # asymptomatic infected humans
        I_h,  # symptomatic infected humans
        R_h,  # recovered humans
        S_v,  # susceptible mosquitoes
        E_v,  # exposed mosquitoes
        I_v,  # infected mosquitoes
        N_h,  # total number of humans
        N_v,  # total number of mosquitoes
    ) = state

    ## Human dynamics ##
    newly_susceptible = (
        config.lambda_h  # recruitment rate
        + config.recovery_rate * (1.0 - config.nu) * I_h  # recovery without immunity
        + config.varphi_h * R_h  # loss of immunity after recovery
    )

    exposure_rate = (
        config.mosquito_human_infection_rate * I_v * S_h / N_h
        + config.asymptomatic_infection_rate * A_h * S_h / N_h
        + config.symptomatic_infection_rate * I_h * S_h / N_h
    )

    dS_h = newly_susceptible - exposure_rate - config.mu_h * S_h

    # Decrease due to exposed becoming infected and natural human death
    dA_h = exposure_rate - config.alpha_h * A_h - config.mu_h * A_h

    # Increase to due exposed becoming infected and decrease due to recovery,
    # natural human death, and death from Zika
    dI_h = (
        config.alpha_h * A_h
        - (config.recovery_rate + config.mu_h + config.delta_h) * I_h
    )

    # Increase due to recovery with immunity and decrease due to death
    # and gradual loss of immunity after recovery
    dR_h = (
        config.recovery_rate * config.nu * I_h - (config.varphi_h + config.mu_h) * R_h
    )

    dN_h = config.lambda_h - config.delta_h * I_h - config.mu_h * N_h

    ## Mosquito dynamics ##
    dS_v = (
        config.lambda_v  # Recruiment rate
        # Newly exposed mosquitos
        - config.human_mosquito_infection_rate * I_h * S_v / N_h
        - config.mu_v * S_v  # Natural death rate
        - config.theta * config.indoor_spray_use * S_v  # Death due to indoor spray
    )

    dE_v = (
        # Newly exposed mosquitos
        config.human_mosquito_infection_rate * I_h * S_v / N_h
        - config.alpha_v * E_v  # Newly infected mosquitos
        - config.mu_v * E_v  # Natural death rate
        - config.theta * config.indoor_spray_use * E_v  # Death due to indoor spray
    )

    dI_v = (
        config.alpha_v * E_v  # Newly infected mosquitos
        - config.mu_v * I_v  # Natural death rate
        - config.theta * config.indoor_spray_use * I_v  # Death due to indoor spray
    )

    # Increase due to recruitment and decrease is due to natural mosquito death
    # and death due to indoor spray use.
    dN_v = (
        config.lambda_v - (config.mu_v + config.theta * config.indoor_spray_use) * N_v
    )

    ds_dt = [
        dS_h,
        dA_h,
        dI_h,
        dR_h,
        dS_v,
        dE_v,
        dI_v,
        dN_h,
        dN_v,
    ]
    return ds_dt


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the Zika simulator model.

    The simulation starts at initial_state at time 0, and evolves the state
    using dynamics whose parameters are specified in config.

    Parameters
    ----------
        initial_state:  `whynot.simulators.zika.State`
            Initial State object, which is used as x_{t_0} for the simulator.
        config:  `whynot.simulators.zika.Config`
            Config object that encapsulates the parameters that define the dynamics.
        intervention: `whynot.simulators.zika.Intervention`
            Intervention object that specifies what, if any, intervention to perform.
        seed: int
            Seed to set internal randomness. The simulator is deterministic, so
            the seed parameter is ignored.

    Returns
    -------
        run: `whynot.dynamics.Run`
            Rollout of the model.

    """
    # Simulator is deterministic, so seed is ignored
    # pylint: disable-msg=unused-argument
    t_eval = np.arange(
        config.start_time, config.end_time + config.delta_t, config.delta_t
    )

    solution = odeint(
        dynamics,
        y0=dataclasses.astuple(initial_state),
        t=t_eval,
        args=(config, intervention),
        rtol=config.rtol,
        atol=config.atol,
    )

    states = [initial_state] + [State(*state) for state in solution[1:]]
    return wn.dynamics.Run(states=states, times=t_eval)


if __name__ == "__main__":
    print(State())
    print(simulate(State(), Config()))



"""Credit simulator initialization."""

from whynot.simulators.credit.simulator import (
    agent_model,
    Config,
    dynamics,
    Intervention,
    simulate,
    strategic_logistic_loss,
    State,
)

from whynot.simulators.credit.dataloader import CreditData

from whynot.simulators.credit.environments import *

SUPPORTS_CAUSAL_GRAPHS = True



"""Ordinary least-squares based-estimators for causal inference."""
from time import perf_counter

import numpy as np
import statsmodels.api as sm

from whynot.framework import InferenceResult


def estimate_treatment_effect(covariates, treatment, outcome):
    """Run ordinary least squares to estimate causal effect of treatment variable on the outcome Y.

    Parameters
    ----------
        covariates: `np.ndarray`
            Array of shape [num_samples, num_features] of features
        treatment:  `np.ndarray`
            Binary array of shape [num_samples]  indicating treatment status for each
            sample.
        outcome:  `np.ndarray`
            Array of shape [num_samples] containing the observed outcome for each sample.

    Returns
    -------
        result: `whynot.framework.InferenceResult`
            InferenceResult object for this procedure

    """
    features = np.copy(covariates)
    treatment = treatment.reshape(-1, 1)
    features = np.concatenate([treatment, features], axis=1)
    features = sm.add_constant(features, prepend=True, has_constant="add")

    # Only time model fitting, not preprocessing
    start_time = perf_counter()
    model = sm.OLS(outcome, features)
    results = model.fit()
    stop_time = perf_counter()

    # Treatment is the second variable (first is the constant offset)
    ate = results.params[1]
    stderr = results.bse[1]
    conf_int = (ate - 1.96 * stderr, ate + 1.96 * stderr)
    return InferenceResult(
        ate=ate,
        stderr=stderr,
        ci=conf_int,
        individual_effects=None,
        elapsed_time=stop_time - start_time,
    )



"""Simple integration and determinism tests for all simulators."""
import numpy as np
import pytest

import whynot as wn


def check_shapes(dataset, num_samples):
    """Ensure datast has the right size."""
    assert (
        len(dataset.covariates.shape) == 2
        and dataset.covariates.shape[0] == num_samples
    )
    assert dataset.treatments.shape == (num_samples,)
    assert dataset.outcomes.shape == (num_samples,)
    assert dataset.true_effects.shape == (num_samples,)


@pytest.mark.parametrize(
    "simulator",
    [
        wn.delayed_impact,
        wn.hiv,
        wn.lotka_volterra,
        wn.opioid,
        wn.world2,
        wn.world3,
        wn.zika,
    ],
    ids=[
        "delayed_impact",
        "hiv",
        "lotka_volterra",
        "opioid",
        "world2",
        "world3",
        "zika",
    ],
)
def test_dynamics_initial_state(simulator):
    """For ODE simulators, ensure the iniitial_state is returned by reference in run."""
    initial_state = simulator.State()
    config = simulator.Config()
    run = simulator.simulate(initial_state, config)
    assert len(run.states) == len(run.times)
    assert run.initial_state is initial_state
    assert np.allclose(config.delta_t, run.times[1] - run.times[0])


@pytest.mark.parametrize(
    "simulator,intervention_param,intervention_val",
    [
        (wn.delayed_impact, "credit_scorer", lambda score: min(score, 600)),
        (wn.hiv, None, None),
        (wn.lotka_volterra, None, None),
        (wn.opioid, "nonmedical_incidence", -0.1),
        (wn.world2, None, None),
        (wn.world3, None, None),
        (wn.zika, "treated_bednet_use", 0.5),
    ],
    ids=[
        "delayed_impact",
        "hiv",
        "lotka_volterra",
        "opioid",
        "world2",
        "world3",
        "zika",
    ],
)
def test_dynamics_intervention(simulator, intervention_param, intervention_val):
    """For ODE simulators, ensure test config.intervention."""
    initial_state = simulator.State()
    config = simulator.Config(delta_t=1.0)
    intervention_time = (config.start_time + config.end_time) // 2

    if intervention_param is None:
        intervention_param = config.parameter_names()[0]
        intervention_val = 5.0 * getattr(config, intervention_param)

    kwargs = {intervention_param: intervention_val}
    intervention = simulator.Intervention(time=intervention_time, **kwargs)

    # Run the simulator with and without the intervention
    # Ensure the states match up to the intervention time and diverge
    # thereafter.
    untreated_run = simulator.simulate(
        initial_state, config, intervention=None, seed=1234
    )
    treated_run = simulator.simulate(
        initial_state, config, intervention=intervention, seed=1234
    )

    for idx, time in enumerate(untreated_run.times):
        if time < intervention_time:
            assert np.allclose(
                untreated_run.states[idx].values(), treated_run.states[idx].values()
            )
        elif time >= intervention_time + config.delta_t:
            # Should make sure the interventions have *some* effect
            assert not np.allclose(
                untreated_run.states[idx].values(), treated_run.states[idx].values()
            )


@pytest.mark.parametrize(
    "simulator,num_samples",
    [
        (wn.civil_violence, 5),
        (wn.delayed_impact, 10),
        (wn.dice, 10),
        (wn.hiv, 10),
        (wn.lalonde, 445),
        (wn.lotka_volterra, 10),
        (wn.opioid, 10),
        (wn.schelling, 5),
        (wn.world2, 10),
        (wn.world3, 10),
        (wn.zika, 10),
    ],
    ids=[
        "civil_violence",
        "delayed_impact",
        "dice",
        "hiv",
        "lalonde",
        "lotka_volterra",
        "opioid",
        "schelling",
        "world2",
        "world3",
        "zika",
    ],
)
def test_simulator_experiments(simulator, num_samples):
    """Test simulator experiments for determinism and proper outcome size."""
    for experiment in simulator.get_experiments():
        params = experiment.get_parameters()
        print(params)
        sampled_params = params.sample(seed=1234)
        dataset1 = experiment.run(num_samples=num_samples, seed=1234, **sampled_params)

        # Sanity check to make sure data is well-formed.
        check_shapes(dataset1, num_samples)

        # Rerun the experiment for determinism.
        dataset2 = experiment.run(num_samples=num_samples, seed=1234, **sampled_params)
        assert np.allclose(dataset1.covariates, dataset2.covariates)
        assert np.allclose(dataset1.treatments, dataset2.treatments)
        assert np.allclose(dataset1.outcomes, dataset2.outcomes)
        assert np.allclose(dataset1.true_effects, dataset2.true_effects)


@pytest.mark.parametrize(
    "simulator,num_samples",
    [
        (wn.civil_violence, 5),
        (wn.delayed_impact, 10),
        (wn.dice, 10),
        (wn.hiv, 10),
        (wn.lalonde, 445),
        (wn.lotka_volterra, 10),
        (wn.opioid, 10),
        (wn.schelling, 5),
        (wn.world2, 10),
        (wn.world3, 10),
        (wn.zika, 10),
    ],
    ids=[
        "civil_violence",
        "delayed_impact",
        "dice",
        "hiv",
        "lalonde",
        "lotka_volterra",
        "opioid",
        "schelling",
        "world2",
        "world3",
        "zika",
    ],
)
def test_parallelize(simulator, num_samples):
    """Test that parallelized runs are identical to non-parallelized."""
    for experiment in simulator.get_experiments():
        unparallelized = experiment.run(
            num_samples=num_samples, parallelize=False, seed=1234
        )
        parallelized = experiment.run(
            num_samples=num_samples, parallelize=True, seed=1234
        )

        assert np.allclose(unparallelized.covariates, parallelized.covariates)
        assert np.allclose(unparallelized.treatments, parallelized.treatments)
        assert np.allclose(unparallelized.outcomes, parallelized.outcomes)
        assert np.allclose(unparallelized.true_effects, parallelized.true_effects)



"""Implements the classic Schelling model in the Mesa framework."""

from mesa import Model, Agent
from mesa.time import RandomActivation
from mesa.space import SingleGrid
from mesa.datacollection import DataCollector


class SchellingAgent(Agent):
    """Schelling segregation agent model."""

    def __init__(self, pos, model, agent_type, homophily):
        """Create a new Schelling agent.

        Parameters
        ----------
            unique_id: Unique identifier for the agent.
            x, y: Agent initial location.
            agent_type: Indicator for the agent's type (minority=1, majority=0)

        """
        super().__init__(pos, model)
        self.pos = pos
        self.type = agent_type
        self.homophily = homophily

    def step(self):
        """Update the agent model after one step."""
        similar = 0
        for neighbor in self.model.grid.neighbor_iter(self.pos):
            if neighbor.type == self.type:
                similar += 1

        # If unhappy, move:
        if similar < self.homophily:
            self.model.grid.move_to_empty(self)
        else:
            self.model.happy += 1


class Schelling(Model):
    """Model class for the Schelling segregation model."""

    def __init__(
        self,
        height=20,
        width=20,
        density=0.8,
        minority_pc=0.2,
        homophily=3,
        education_boost=0,
        education_pc=0.2,
        seed=None,
    ):
        """Seed is used to set randomness in the __new__ function of the Model superclass."""
        # pylint: disable-msg=unused-argument,super-init-not-called

        self.height = height
        self.width = width
        self.density = density
        self.minority_pc = minority_pc
        self.homophily = homophily
        self.education_boost = education_boost
        self.education_pc = education_pc

        self.schedule = RandomActivation(self)
        self.grid = SingleGrid(height, width, torus=True)

        self.happy = 0
        self.datacollector = DataCollector(
            {"happy": "happy"},  # Model-level count of happy agents
            # For testing purposes, agent's individual x and y
            {"x": lambda a: a.pos[0], "y": lambda a: a.pos[1]},
        )

        # Set up agents
        # We use a grid iterator that returns
        # the coordinates of a cell as well as
        # its contents. (coord_iter)
        for cell in self.grid.coord_iter():
            x_coord = cell[1]
            y_coord = cell[2]
            if self.random.random() < self.density:
                if self.random.random() < self.minority_pc:
                    agent_type = 1
                else:
                    agent_type = 0

                agent_homophily = homophily
                if self.random.random() < self.education_pc:
                    agent_homophily += self.education_boost

                agent = SchellingAgent(
                    (x_coord, y_coord), self, agent_type, agent_homophily
                )
                self.grid.position_agent(agent, (x_coord, y_coord))
                self.schedule.add(agent)

        self.running = True
        self.datacollector.collect(self)

    def step(self):
        """Run one step of the model. If All agents are happy, halt the model."""
        self.happy = 0  # Reset counter of happy agents
        self.schedule.step()
        # collect data
        self.datacollector.collect(self)

        if self.happy == self.schedule.get_agent_count():
            self.running = False



"""Ensure gym spaces are accessible if you import whynot.gym as gym."""
from gym.spaces.space import Space
from gym.spaces.box import Box
from gym.spaces.discrete import Discrete
from gym.spaces.multi_discrete import MultiDiscrete
from gym.spaces.multi_binary import MultiBinary
from gym.spaces.tuple import Tuple
from gym.spaces.dict import Dict

from gym.spaces.utils import flatdim
from gym.spaces.utils import flatten
from gym.spaces.utils import unflatten

__all__ = [
    "Space",
    "Box",
    "Discrete",
    "MultiDiscrete",
    "MultiBinary",
    "Tuple",
    "Dict",
    "flatdim",
    "flatten",
    "unflatten",
]



"""Incarceration simulator.

Agent-based model for incarceration dynamics. The simulator models incarceration
as an infectious property that can be passed on along social ties. Infection
probabilities are based on survey data. The simulator is based on Lum, Swarup,
Eubank, and Hawdon (2014). URL: https://doi.org/10.1098/rsif.2014.0409

The simulation happens on top of a fixed set of agents provided by the authors
of the study. The number of agents cannot be varied.

"""

import os
import dataclasses
import random
import gzip
import pickle

import numpy as np
from tqdm.auto import tqdm


@dataclasses.dataclass
class Config:
    # pylint: disable-msg=too-few-public-methods
    """Metaparameters for a single run of simulator.

    Attributes
    ----------
    random_seed : int
        Controls all randomness of the simulation
    min_age : int
        Minimum age for incarceration
    start_iter : int
        Start iteration of simulation
    end_iter : int
        End iteration of simulation
    percent : float
        Initial incarceration rate
    random_sentence_type : bool
        Assign lenient or harsh sentence randomly
    random_sentence_bias : float
        Probability of picking harsh sentence
    mean_sentence_harsh : float
        Mean sentence length (months) for harsh assignment
    mean_sentence_lenient : float
        Mean sentence length (months) for lenient assignment

    """

    random_seed: int = 1
    min_age: int = 15
    start_iter: int = 100
    end_iter: int = 200
    percent: float = 0.01
    harsh_sentence: bool = False
    random_sentence_type: bool = False
    random_sentence_bias: float = 0.5
    consistent_sentence_length: bool = True
    mean_sentence_harsh: float = 17.0
    mean_sentence_lenient: float = 14.0


def infect(person_sex, relation_type, relation_sex):
    """Generate infection based on relation characteristics.

    Parameters
    ----------
    person_sex : string
        Sex of person ('m' or 'f')
    relation_type : string
        Type of social relation  ('parent', 'sibling', 'partner', 'child')
    relation_sex : string
        Sex of related person ('m' or 'f')

    Returns
    -------
    infected : int
        1 if infected, 0 otherwise

    """
    infection_probability_month = {
        "f": {
            "parent": {"f": 0.000849988733768181, "m": 0.0112878570024662,},
            "sibling": {"f": 0.00801193753900653, "m": 0.0332053842229949,},
            "partner": {"*": 0.0043472740358963},
            "child": {"*": 0.0169602401420906},
        },
        "m": {
            "parent": {"f": 0.00347339838166261, "m": 0.0113344842544054,},
            "sibling": {"f": 0.00436688659218365, "m": 0.0301729987453868,},
            "partner": {"*": 0.00078339990345766},
            "child": {"*": 0.00634223110220566},
        },
        "*": {"*": {"*": 1.675041946602729e-05}},
    }

    probability = infection_probability_month[person_sex][relation_type][relation_sex]
    return np.random.binomial(1, probability)


def valid_age(person, i, min_age):
    """Verify age requirement of person in given iteration."""
    return person["birth"] <= (i - min_age) and person["death"] >= i


def initialize(config):
    """Load population and generate initial incarceration state."""
    random.seed(config.random_seed)

    population_pickle = os.path.join(os.path.dirname(__file__), "population.pkl.gz")
    popu = pickle.load(gzip.open(population_pickle, "rb"))

    alive = []
    for num, person in popu.items():
        person["num"] = num
        person["months_in_prison"] = 0
        person["harsh_sentence"] = []
        if valid_age(person, config.start_iter, config.min_age):
            alive.append(person)
    num_infect = round(config.percent * len(alive))
    potentials = []
    for person in alive:
        if config.start_iter - person["birth"] <= 45:
            potentials.append(person)
    infected = random.sample(potentials, num_infect)
    for person in infected:
        sentence, harsh_sentence = generate_sentence(person, -1, 0, config)
        person["incarcerated"] = sentence
        person["harsh_sentence"].append(harsh_sentence)

    return popu


def spread_infection(popu, person, itr, month, config):
    """Pass on infection."""
    # ensure same infection patterns regardless of sentence intervention
    np.random.seed(
        hash((config.random_seed, person["num"], itr, month)) % (2 ** 32 - 1)
    )

    sex = person["sex"]

    for sibling in person["siblings"]:
        popu[sibling]["infected"] += infect(sex, "sibling", popu[sibling]["sex"])

    if person["partner"] >= 0 and person["iter_married"] >= itr:
        partner = person["partner"]
        popu[partner]["infected"] += infect(popu[partner]["sex"], "partner", "*")

    for child in person["children"]:
        if valid_age(popu[child], itr, config.min_age):
            popu[child]["infected"] += infect(sex, "child", "*")

    for friend in person["friends"]:
        popu[friend]["infected"] += infect(sex, "sibling", popu[friend]["sex"])

    for parent in person["parents"]:
        popu[parent]["infected"] += infect(sex, "parent", popu[parent]["sex"])


def generate_sentence(person, itr, month, config):
    """Generate sentence based on harshness setting."""
    # generate both sentences to ensure consistent counterfactuals
    # offset random seed to avoid interference with infection random seed
    np.random.seed(
        hash((1, config.random_seed, person["num"], itr, month)) % (2 ** 32 - 1)
    )

    gamma_parameter = 1.2

    lenient_mean = config.mean_sentence_lenient
    gamma_sample = np.random.gamma(gamma_parameter, lenient_mean / gamma_parameter)
    lenient_sentence = np.random.poisson(gamma_sample)

    if config.consistent_sentence_length:
        penalty = config.mean_sentence_harsh - config.mean_sentence_lenient
        harsh_sentence = lenient_sentence + penalty
    else:
        harsh_mean = config.mean_sentence_harsh
        gamma_sample = np.random.gamma(gamma_parameter, harsh_mean / gamma_parameter)
        harsh_sentence = np.random.poisson(gamma_sample)

    if config.random_sentence_type:
        if np.random.binomial(1, config.random_sentence_bias):
            return harsh_sentence, True
        return lenient_sentence, False

    if config.harsh_sentence:
        return harsh_sentence, True
    return lenient_sentence, False


def assign_sentence(person, itr, month, config):
    """Assign sentence."""
    if valid_age(person, itr, config.min_age):
        sentence, harsh_sentence = generate_sentence(person, itr, month, config)
        person["incarcerated"] = sentence
        person["harsh_sentence"].append(harsh_sentence)


def simulate(config, show_progress=False):
    """Simulate incarceration contagion dynamics.

    Parameters
    ----------
    config : Config
        Config object specifying simulation parameters.

    Returns
    -------
    dict
        Dictionary specifying simulated population of agents.

    """
    popu = initialize(config)

    agents = popu.values()

    def display(range_obj):
        if show_progress:
            range_obj = tqdm(range_obj)
        return range_obj

    # these are in years. need to work in terms of months
    for itr in display(range(config.start_iter, config.end_iter)):

        for month in range(12):

            # infection step
            for person in agents:

                # random infection, not due to contagion
                if valid_age(person, itr, config.min_age):
                    person["infected"] += infect("*", "*", "*")

                # infect connected people
                if person["incarcerated"] > 0:
                    person["incarcerated"] -= 1
                    person["months_in_prison"] += 1
                    spread_infection(popu, person, itr, month, config)

            # sentencing step
            for person in agents:
                if person["infected"] and not person["incarcerated"]:
                    assign_sentence(person, itr, month, config)
                person["infected"] = 0

    return popu



"""Unit tests for setup and intervention in delayed impact simulator."""
import numpy as np
import statsmodels

from whynot.simulators.delayed_impact.simulator import *


def test_lending_policy():
    config = Config(threshold_g0=600, threshold_g1=650)
    assert lending_policy(config, group=0, score=599) == 0
    assert lending_policy(config, group=0, score=600) == 1
    assert lending_policy(config, group=0, score=660) == 1
    assert lending_policy(config, group=1, score=649) == 0
    assert lending_policy(config, group=1, score=650) == 1
    assert lending_policy(config, group=1, score=660) == 1


def test_update_score():
    config = Config(
        repayment_score_change=50,
        default_score_change=-80,
        min_score=300,
        max_score=800,
    )

    score = 600
    # No change if loan not approved
    assert update_score(config, score, loan_approved=False, repaid=True) == score
    assert update_score(config, score, loan_approved=False, repaid=False) == score
    assert update_score(config, score, loan_approved=True, repaid=True) == score + 50
    assert update_score(config, score, loan_approved=True, repaid=False) == score - 80

    score = 325
    assert update_score(config, score, loan_approved=True, repaid=False) == 300

    score = 780
    assert update_score(config, score, loan_approved=True, repaid=True) == 800


def test_update_profits():
    config = Config(repayment_utility=10, default_utility=-40)

    profits = 10
    assert (
        update_profits(config, profits, loan_approved=True, repaid=False)
        == profits - 40
    )
    assert (
        update_profits(config, profits, loan_approved=True, repaid=True) == profits + 10
    )
    assert update_profits(config, profits, loan_approved=False, repaid=True) == profits
    assert update_profits(config, profits, loan_approved=False, repaid=False) == profits


def check_repayment_rate(repayments, prob):
    """Ensure that repayment has approximately the correct proportion of repaid."""
    total_repaid = np.sum(repayments)
    ci_low, ci_upp = statsmodels.stats.proportion.proportion_confint(
        count=total_repaid, nobs=len(repayments), alpha=0.001, method="beta"
    )
    assert ci_low <= prob <= ci_upp


def test_repayment_fcn():
    rng = np.random.RandomState(1234)
    score = 600

    # Group 0
    for group in range(2):
        repayments = [determine_repayment(rng, group, score) for _ in range(500)]
        check_repayment_rate(repayments, LOAN_REPAY_PROBS[group](score))



"""Simulator for Nordhaus' 2007 DICE model."""
from contextlib import contextmanager
import copy
import dataclasses
import os
import sys

import numpy as np
import pyomo
from pyomo.environ import (
    RangeSet,
    ConcreteModel,
    NonNegativeReals,
    Reals,
    Var,
    Constraint,
    Objective,
    summation,
    log,
    maximize,
)

import whynot as wn
from whynot.dynamics import BaseConfig, BaseIntervention, BaseState


@dataclasses.dataclass
class Config(BaseConfig):
    """Parameter values in the DICE model.

    Default values correspond to the base run of the 2013 version.
    """

    # pylint:disable-msg=too-many-public-methods,invalid-name
    # pylint:disable-msg=too-many-instance-attributes
    #: Number of time periods to run the simulation.
    numPeriods: int = 60
    #: Number of year for each time period.
    tstep: int = 5

    # Preferences
    #: Elasticity of marginal utility of consumption.
    elasmu: float = 1.45
    #: Initial rate of social time preference per year.
    prstp: float = 0.015

    # Population and technology
    #: Capital elasticity in production function.
    gama: float = 0.3
    #: Initial world population (millions)
    pop0: float = 6838
    #: Growth rate to calibrate 2050 population projection.
    popadj: float = 0.134
    #: Asymptotic population (millions)
    popasym: float = 10500
    #: Depreciation rate on capital (per year)
    dk: float = 0.1
    #: Initial world gross output (trill 2005 USD)
    q0: float = 63.69
    #: Initial capital value (trill 2005 USD)
    k0: float = 135
    #: Initial level of total factor productivity
    a0: float = 3.8
    #: Initial growth rate for total factor productivity (per 5 years)
    ga0: float = 0.079
    #: Decline rate of total factor productivity (per 5 years)
    dela: float = 0.006

    # Emissions
    #: Initial growth rate of sigma (per year)
    gsigma1: float = -0.01
    #: Decline rate of decarbonization (per period)
    dsig: float = -0.001
    #: Carbon emissions from land 2010 (GtCO2 per year)
    eland0: float = 3.3
    #: Decline rate of land emissions (per period)
    deland: float = 0.2
    #: Industrial emissions 2010 (GtC02 per year)
    e0: float = 33.61
    #: Initial emissions control rate for base case 2010
    miu0: float = 0.039

    # Carbon cycle
    #: Initial concentration in atmosphere 2010 (GtC)
    mat0: float = 830.4
    #: Initial concentration in upper strata 2010 (GtC)
    mu0: float = 1527
    #: Initial concentration in lower strata 2010 (GtC)
    ml0: float = 10010
    #: Equilibrium concentration atmosphere (GtC)
    mateq: float = 588
    #: Equilibrium concentration in upper strata (GtC)
    mueq: float = 1350
    #: Equilibrium concentration in lower strata (GtC)
    mleq: float = 10000
    #: Carbon cycle transition matrix.
    b12: float = 0.088
    #: Carbon cycle transition matrix.
    b23: float = 0.0025

    # Climate model
    #: Equilibrium temperature impact (oC per doubling CO2)
    t2xco2: float = 2.9
    #: 2010 forcings of non-CO2 CHG (Wm-2)
    fex0: float = 0.25
    #: 2100 forcings of non-CO2 CHG (Wm-2)
    fex1: float = 0.7
    #: Initial lower stratum temperature change (C from 1900)
    tocean0: float = 0.0068
    #: Initial atmospheric temperature change (C from 1900)
    tatm0: float = 0.8
    #: Climate equation coefficient for upper level
    c1: float = 0.098
    #: Tranfer coefficient upper to lower stratum
    c3: float = 0.088
    #: Transfer coefficient for lower level
    c4: float = 0.025
    #: Forcings of equilibrium CO2 doubling (Wm-2)
    fco22x: float = 3.8

    # Climate damage parameters calibrates for quadratic at 2.5 C for 2105
    #: Damage intercept
    a1: float = 0
    #: Damage quadratic term
    a2: float = 0.00267
    #: Damage exponent
    a3: float = 2

    # Abatement cost
    #: Exponent of control cost function
    expcost2: float = 2.8
    #: Cost of backstop 2005$ for tCO2 2010
    pback: float = 344
    #: Initial cost decline backstop cost per period
    gback: float = 0.025
    #: Upper limit on control rate after 2150
    limmiu: float = 1.2
    #: Period before which no emissions controls base
    tnopol: float = 45
    #: Initial base carbon price (2005$ per tCO2)
    cprice0: float = 1
    #: Growth rate of base carbon price per year.
    gcprice: float = 0.02
    #: Period at which to have full participation.
    periodfullpart: int = 21
    #: Fraction of emissions under control in 2010
    partfract2010: float = 1
    #: Fraction of emissions under control at full time
    partfractfull: float = 1

    # Availability of fossil fuels
    #: Maximum cumulative extraction fossil fuels (GtC)
    fosslim: float = 6000

    # Scaling and inessential parameters
    #: Multiplicative scaling coefficient
    scale1: float = 0.016408662
    #: Additive scaling coefficient
    scale2: float = -3855.106895

    #: Whether or not to use optimization to set the carbon price.
    ifopt: int = 1

    # Parameters for long-run consistency of carbon cycle
    @property
    def b11(self):
        """Carbon cycle transition matrix."""
        return 1 - self.b12

    @property
    def b21(self):
        """Carbon cycle transition matrix."""
        return self.b12 * self.mateq / self.mueq

    @property
    def b22(self):
        """Carbon cycle transition matrix."""
        return 1 - self.b21 - self.b23

    @property
    def b32(self):
        """Carbon cycle transition matrix."""
        return self.b23 * self.mueq / self.mleq

    @property
    def b33(self):
        """Carbon cycle transition matrix."""
        return 1 - self.b32

    @property
    def sig0(self):
        """Carbon intensity 2010 (kgCO2 per output 2005 USD 2010)."""
        return self.e0 / (self.q0 * (1 - self.miu0))

    @property
    def lam(self):
        """Climate model parameter."""
        return self.fco22x / self.t2xco2

    @property
    def L(self):  # pylint:disable-msg=invalid-name
        """Level of population and labor."""
        pop = {1: self.pop0}
        for time in range(1, self.numPeriods):
            pop[time + 1] = pop[time] * ((self.popasym / pop[time]) ** self.popadj)
        return pop

    def ga(self, time):  # pylint:disable-msg=invalid-name
        """Growth rate of productivity."""
        return self.ga0 * np.exp(-self.dela * 5.0 * (time - 1))

    def al(self, time):  # pylint:disable-msg=invalid-name
        """Level of total factor productivity."""
        if time == 1:
            return self.a0
        return self.al(time - 1) / (1 - self.ga(time - 1))

    def gsig(self, time):
        """Change in sigma (cumulative improvement of energy efficiency)."""
        if time == 1:
            return self.gsigma1
        return self.gsig(time - 1) * ((1 + self.dsig) ** self.tstep)

    def sigma(self, time):
        """CO2-equivalent-emissions output ratio."""
        if time == 1:
            return self.sig0
        return self.sigma(time - 1) * np.exp(self.gsig(time - 1) * self.tstep)

    def pbacktime(self, time):
        """Backstop price."""
        return self.pback * (1 - self.gback) ** (time - 1)

    def cost1(self, time):
        """Cost adjusted for backstop."""
        return self.pbacktime(time) * self.sigma(time) / self.expcost2 / 1000.0

    def etree(self, time):
        """Emissions from deforestation."""
        return self.eland0 * (1 - self.deland) ** (time - 1)

    def rr(self, time):  # pylint:disable-msg=invalid-name
        """Average utility social discount rate."""
        return 1.0 / ((1 + self.prstp) ** (self.tstep * (time - 1)))

    def forcoth(self, time):
        """Exogenous forcing for other greenhouse gases."""
        if time < 19:
            return self.fex0 + (1 / 18) * (self.fex1 - self.fex0) * (time)
        return self.fex0 + (self.fex1 - self.fex0)

    @property
    def optlrsav(self):
        """Optimal long-run savings rate used for transversality."""
        return (
            (self.dk + 0.004) / (self.dk + 0.004 * self.elasmu + self.prstp) * self.gama
        )

    def partfract(self, time):
        """Fraction of emissions in control regime."""
        if time == 1:
            return self.partfract2010

        if time > self.periodfullpart:
            return self.partfractfull

        return (
            self.partfract2010
            + (self.partfractfull - self.partfract2010)
            * (time - 1)
            / self.periodfullpart
        )

    def cpricebase(self, time):
        """Carbon price in base case."""
        return self.cprice0 * (1 + self.gcprice) ** (5 * (time - 1))

    def update(self, intervention):
        """Generate a new config object after applying the intervention."""
        return dataclasses.replace(self, **intervention.updates)


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Encapsulate an intervention in DICE."""

    def __init__(self, **kwargs):
        """Construct an intervention in DICE.

        Currently, all interventions are performed at the first time step.

        Parameters
        ----------
            kwargs: dict
                Only valid keyword arguments are parameters of
                `whynot.simulators.dice.Config`.

        """
        super(Intervention, self).__init__(Config, time=0, **kwargs)


@dataclasses.dataclass
class State(BaseState):  # pylint:disable-msg=too-few-public-methods
    """State variables of the DICE simulator.

    Default values are extracted from the first time step of a run of the
    DICE model using optimization to set the carbon price.

    """

    # pylint:disable-msg=invalid-name
    # pylint:disable-msg=too-many-instance-attributes
    #: Emission control rate GHGs
    MIU: float = 0.038976322
    #: Radiative forcing in watts per m2
    FORC: float = 2.167363097
    #: Temperature of atmosphere in degrees C
    TATM: float = 0.8
    #: Temperature of lower oceans in degrees C
    TOCEAN: float = 0.0068
    #: Carbon concentration in atmosphere GtC
    MAT: float = 830.4
    #: Carbon concentration in shallow oceans GtC
    MU: float = 1527
    #: Carbon concentration in lower oceans GtC
    ML: float = 10010
    #: CO2-equivalent emissions GtC
    E: float = 36.85382682
    EIND: float = 33.55382682
    #: Consumption trillions US dollars
    C: float = 46.98638871
    #: Capital stock trillions US dollars
    K: float = 135
    #: Per capita consumption thousands US dollars
    CPC: float = 6.871364246
    #: Investment trillions US dollars
    I: float = 16.48646319
    #: Gross savings rate as fraction of gross world product
    S: float = 0.259740388
    #: Real interest rate per annum
    RI: float = 0.052124994
    #: Gross world product net of abatement and damages
    Y: float = 63.4728519
    #: Gross world product Gross of abatement and damages
    YGROSS: float = 63.58198682
    #: Output net damages equation
    YNET: float = 63.47333792
    #: Damages (trillions 2005 USD per year)
    DAMAGES: float = 0.108648899
    #: Damages as fraction of gross output
    DAMFRAC: float = 0.0017088
    #: Cost of emissions reductions
    ABATECOST: float = 0.000486016
    #: Marginal cost of abatement (2005$ per ton CO2)
    MCABATE: float = 0.999999958
    #: Cumulative industrial carbon emissions GtC
    CCA: float = 90
    #: One period utility function
    PERIODU: float = 0.288713996
    #: Carbon price (2005$ per ton of CO2)
    CPRICE: float = 0.999999958
    #: Period utility
    CEMUTOTPER: float = 1974.226305

    @property
    def variables(self):
        """Return names of all model variables."""
        return [var.name for var in dataclasses.fields(self)]

    @property
    def nonnegative_variables(self):
        """Return names of all nonnegative variables."""
        return ["MIU", "TATM", "MAT", "MU", "ML", "Y", "YGROSS", "C", "K", "I"]

    def __getitem__(self, var_name):
        """Look up variables via state[name]."""
        return getattr(self, var_name)

    def __setitem__(self, name, value):
        """Set state variables via state[name] = value."""
        setattr(self, name, value)


def add_constraint(model, func):
    """Add a time-varying constraint to pyomo model."""
    constraint_name = func.__name__
    model.add_component(constraint_name, Constraint(model.time, rule=func))


def get_ipopt_solver():
    """Construct a platform specific IPOPT solver."""
    cur_dir = os.path.abspath(os.path.dirname(__file__))
    if sys.platform == "linux":
        executable_path = os.path.join(cur_dir, "ipopt_bin", "ipopt-linux64")
    elif sys.platform == "darwin":
        executable_path = os.path.join(cur_dir, "ipopt_bin", "ipopt-osx")
    else:
        raise ValueError(f"No IPOPT binaries for platform {sys.platform}")
    return pyomo.opt.SolverFactory("ipopt", executable=executable_path)


def initialize_model(model, initial_state, config):
    """Create model variables and set the initial state."""
    var_bounds = generate_bounds(config)

    # Create and initialize model variables
    for var_name in initial_state.variables:
        domain = (
            NonNegativeReals
            if (var_name in initial_state.nonnegative_variables)
            else Reals
        )
        bounds = var_bounds.get(var_name, None)
        model.add_component(var_name, Var(model.time, domain=domain, bounds=bounds))

        # Explicitly set the initial variable value only if the initial
        # state doesn't match the default state.
        initial_value = initial_state[var_name]
        default_value = State()[var_name]
        if initial_value != default_value:
            # Clamp the first time step to the initialized value
            getattr(model, var_name)[1].fix(initial_value)


def generate_bounds(config):
    """Generate variable bound functions for initialization."""

    def miu_bounds(_, time):
        if time == 1:
            return (1e-20, None)
        if time < 30:
            return (1e-20, 1)
        return (1e-20, config.limmiu * config.partfract(time))

    def tatm_bounds(_, time):
        if time == 1:
            return (config.tatm0, config.tatm0)
        return (None, 40.0)

    def tocean_bounds(_, time):
        if time == 1:
            return (config.tocean0, config.tocean0)
        return (-1, 20.0)

    def mat_bounds(_, time):
        if time == 1:
            return (config.mat0, config.mat0)
        return (10, None)

    def mu_bounds(_, time):
        if time == 1:
            return (config.mu0, config.mu0)
        return (100.0, None)

    def ml_bounds(_, time):
        if time == 1:
            return (config.ml0, config.ml0)
        return (1000.0, None)

    def c_bounds(*_):
        return (2.0, None)

    def k_bounds(_, time):
        if time == 1:
            return (config.k0, config.k0)
        return (1.0, None)

    def cpc_bounds(*_):
        return (0.01, None)

    def cca_bounds(_, time):
        if time == 1:
            return (90, 90)
        return (None, config.fosslim)

    def s_bounds(_, time):
        if time <= 50:
            return (None, None)
        return (config.optlrsav, config.optlrsav)

    def cprice_bounds(_, time):
        """Use base carbon price if base, otherwise optimized.

        Warning if parameters are changed, the next equation might make base case infeasible
        If so, reduce tnopol so that don't run out of resources.
        """
        if config.ifopt == 0:
            return (None, config.cpricebase(time))
        if time == 1:
            return (None, config.cpricebase(1))
        if time > config.tnopol:
            return (None, 1000)
        return (None, None)

    variable_bounds = {
        "MIU": miu_bounds,
        "TATM": tatm_bounds,
        "TOCEAN": tocean_bounds,
        "MAT": mat_bounds,
        "MU": mu_bounds,
        "ML": ml_bounds,
        "C": c_bounds,
        "K": k_bounds,
        "CPC": cpc_bounds,
        "S": s_bounds,
        "CCA": cca_bounds,
        "CPRICE": cprice_bounds,
    }
    return variable_bounds


def add_emissions_dynamics(model, config):
    """Add emissions and damages constraints."""

    def eeq(model, time):
        """Emissions equation."""
        return model.E[time] == model.EIND[time] + config.etree(time)

    add_constraint(model, eeq)

    def eindeq(model, time):
        """Industrial emissions."""
        return model.EIND[time] == config.sigma(time) * model.YGROSS[time] * (
            1 - (model.MIU[time])
        )

    add_constraint(model, eindeq)

    def ccaeq(model, time):
        """Cumulative carbon emissions."""
        if time < config.numPeriods:
            return model.CCA[time + 1] == model.CCA[time] + model.EIND[time] * 5 / 3.666
        return Constraint.Skip

    add_constraint(model, ccaeq)

    def foreq(model, time):
        return model.FORC[time] == config.fco22x * (
            (log((model.MAT[time] / 588.000)) / log(2))
        ) + config.forcoth(time)

    add_constraint(model, foreq)

    def damfraceq(model, time):
        return model.DAMFRAC[time] == (config.a1 * model.TATM[time]) + (
            config.a2 * model.TATM[time] ** config.a3
        )

    add_constraint(model, damfraceq)

    def dameq(model, time):
        return model.DAMAGES[time] == model.YGROSS[time] * model.DAMFRAC[time]

    add_constraint(model, dameq)

    def abateeq(model, time):
        return model.ABATECOST[time] == model.YGROSS[time] * config.cost1(time) * (
            model.MIU[time] ** config.expcost2
        ) * (config.partfract(time) ** (1 - config.expcost2))

    add_constraint(model, abateeq)

    def mcabateeq(model, time):
        return model.MCABATE[time] == config.pbacktime(time) * model.MIU[time] ** (
            config.expcost2 - 1
        )

    add_constraint(model, mcabateeq)

    def carbpriceeq(model, time):
        return model.CPRICE[time] == config.pbacktime(time) * (
            model.MIU[time] / config.partfract(time)
        ) ** (config.expcost2 - 1)

    add_constraint(model, carbpriceeq)


def add_climate_dynamics(model, config):
    """Impose dynamics of climate and carbon cycle via constraints."""

    def mmat(model, time):
        """Atmospheric concentration equation."""
        if time < config.numPeriods:
            return model.MAT[time + 1] == model.MAT[time] * config.b11 + model.MU[
                time
            ] * config.b21 + (model.E[time] * (5 / 3.666))
        return Constraint.Skip

    add_constraint(model, mmat)

    def mml(model, time):
        """Lower ocean concentration."""
        if time < config.numPeriods:
            return (
                model.ML[time + 1]
                == model.ML[time] * config.b33 + model.MU[time] * config.b23
            )
        return Constraint.Skip

    add_constraint(model, mml)

    def mmu(model, time):
        """Shallow ocean concentration."""
        if time < config.numPeriods:
            return (
                model.MU[time + 1]
                == model.MAT[time] * config.b12
                + model.MU[time] * config.b22
                + model.ML[time] * config.b32
            )
        return Constraint.Skip

    add_constraint(model, mmu)

    def tatmeq(model, time):
        """Temperature-climate equation for atmosphere."""
        if time < config.numPeriods:
            return model.TATM[time + 1] == model.TATM[time] + config.c1 * (
                (
                    model.FORC[time + 1]
                    - (config.fco22x / config.t2xco2) * model.TATM[time]
                )
                - (config.c3 * (model.TATM[time] - model.TOCEAN[time]))
            )
        return Constraint.Skip

    add_constraint(model, tatmeq)

    def toceaneq(model, time):
        """Temperature-climate equation for lower oceans."""
        if time < config.numPeriods:
            return model.TOCEAN[time + 1] == model.TOCEAN[time] + config.c4 * (
                model.TATM[time] - model.TOCEAN[time]
            )
        return Constraint.Skip

    add_constraint(model, toceaneq)


def add_economic_dynamics(model, config):
    """Add dynamics constraints on economic variables."""

    def ygrosseq(model, time):
        return model.YGROSS[time] == (
            config.al(time) * (config.L[time] / 1000) ** (1 - config.gama)
        ) * (model.K[time] ** config.gama)

    add_constraint(model, ygrosseq)

    def yneteq(model, time):
        return model.YNET[time] == model.YGROSS[time] * (1 - model.DAMFRAC[time])

    add_constraint(model, yneteq)

    def yyeq(model, time):
        return model.Y[time] == model.YNET[time] - model.ABATECOST[time]

    add_constraint(model, yyeq)

    def cc(model, time):  # pylint:disable-msg=invalid-name
        return model.C[time] == model.Y[time] - model.I[time]

    add_constraint(model, cc)

    def cpce(model, time):
        return model.CPC[time] == 1000 * model.C[time] / config.L[time]

    add_constraint(model, cpce)

    def seq(model, time):
        return model.I[time] == model.S[time] * model.Y[time]

    add_constraint(model, seq)

    def kkeq(model, time):
        if time < config.numPeriods:
            return (
                model.K[time + 1]
                <= (1 - config.dk) ** config.tstep * model.K[time]
                + config.tstep * model.I[time]
            )
        return Constraint.Skip

    add_constraint(model, kkeq)

    def rieq(model, time):
        if time < config.numPeriods:
            return (
                model.RI[time]
                == (1 + config.prstp)
                * (model.CPC[time + 1] / model.CPC[time])
                ** (config.elasmu / config.tstep)
                - 1
            )
        return Constraint.Skip

    add_constraint(model, rieq)


def add_utility_dynamics(model, config):
    """Add constraints on utility variables."""

    def cemutotpereq(model, time):
        return model.CEMUTOTPER[time] == model.PERIODU[time] * config.L[
            time
        ] * config.rr(time)

    add_constraint(model, cemutotpereq)

    def perideq(model, time):
        return (
            model.PERIODU[time]
            == ((model.C[time] * 1000 / config.L[time]) ** (1 - config.elasmu) - 1)
            / (1 - config.elasmu)
            - 1
        )

    add_constraint(model, perideq)

    def utiliteq(model):
        return (
            model.UTILITY
            == config.tstep * config.scale1 * summation(model.CEMUTOTPER)
            + config.scale2
        )

    model.util = Constraint(rule=utiliteq)


@contextmanager
def silence_stdout():
    """Supress stdout output while active."""
    new_target = open(os.devnull, "w")
    old_target = sys.stdout
    sys.stdout = new_target
    try:
        yield new_target
    finally:
        sys.stdout = old_target


def simulate(initial_state, config, intervention=None, seed=None, stochastic=True):
    """Simulate a run of the DICE model.

    Parameters
    ----------
        initial_state: whynot.simulators.dice.State
            Initial state of for a run of the DICE model
        config: whynot.simulators.dice.Config
            Configuration object to set parameters of the dynamics.
        intervention:  whynot.simulators.dice.Intervention
            (Optional) What, if any, intervention to perform during execution.
        seed: int
            (Optional) Random seed for all model randomness.
        stochastic: bool
            (Optional) Whether or not to apply a random perturbation to the dynamics.

    Returns
    -------
        run: whynot.dynamics.Run
            Sequence of states and corresponding timesteps generated during execution.

    """
    if intervention:
        config = config.update(intervention)

    # Initialize model
    rng = np.random.RandomState(seed)
    model = ConcreteModel()
    model.time = RangeSet(1, config.numPeriods, 1)

    # Generate additional randomness using uncertainty ranges
    # for configuration parameters from the original DICE model.
    if stochastic:
        config = copy.deepcopy(config)
        config.t2xco2 *= rng.uniform(low=0.9, high=1.1)
        config.fosslim *= rng.uniform(low=0.9, high=1.1)
        config.limmiu *= rng.uniform(low=0.9, high=1.1)
        config.dsig *= rng.uniform(low=0.9, high=1.1)
        config.pop0 *= rng.uniform(low=0.9, high=1.1)

    # Initialize model and set initial state
    initialize_model(model, initial_state, config)

    # Add constraints governing model dynamics
    add_emissions_dynamics(model, config)
    add_climate_dynamics(model, config)
    add_economic_dynamics(model, config)

    # Set objective
    model.UTILITY = Var(domain=Reals)
    add_utility_dynamics(model, config)
    model.OBJ = Objective(rule=lambda m: m.UTILITY, sense=maximize)

    # Solve the model
    solver = get_ipopt_solver()
    with silence_stdout():
        _ = solver.solve(
            model,
            tee=True,
            symbolic_solver_labels=True,
            keepfiles=False,
            options={"max_iter": 99900, "halt_on_ampl_error": "yes", "print_level": 0},
        )

    # Read out values from the optimized model
    states, times = [initial_state], [0]
    for time in range(1, config.numPeriods + 1):
        state = State()
        for var in state.variables:
            state[var] = getattr(model, var)[time].value
        states.append(state)
        times.append(time)

    return wn.dynamics.Run(states=states, times=times)


if __name__ == "__main__":
    simulate(State(MAT=600), Config(numPeriods=10), stochastic=False)



"""Experiments for civil violence simulator."""
import copy
import dataclasses

import numpy as np
import pandas as pd

from whynot.framework import GenericExperiment, parameter
from whynot.simulators import civil_violence
from whynot import utils

__all__ = ["get_experiments", "RCT"]


def get_experiments():
    """Return all experiments for civil violence."""
    return [RCT]


def run_simulator(agents, config, treatment_assignment, seed):
    """Run the simulator with the specified agents receiving treatment."""
    population = []
    for agent, treated in zip(agents, treatment_assignment):
        agent = copy.deepcopy(agent)
        if treated:
            agent.risk_aversion = 0.9
        population.append(agent)
    return civil_violence.simulate(population, config, seed)


def run_abm_experiment(
    agents, treatment_assignments, config, rng, parallelize, show_progress=True
):
    """Run an agent-based modeling experiment.

    We run (in parallel) the agent based model for the given treatment assignments to
    simulate a causal experiment. To obtain ground truth counterfactuals, we run
    repeatedly run the simulator with the same random seed where (1) no agent i
    treated [baseline] and (2) where only agent i is treated for i=1, 2, ...
    Thus, the reported ground truth effects are really the marginal causal
    effects.

    Parameters
    ----------
        agents: list
            List of whynot.simulators.civil_violence.Agent
        treatment_assignments: np.ndarray
            Array of shape [num_agents] indicating whether or not each agent is treated.
        config: whynot.simulators.civil_violence.Config
            Simulator parameters to use for each run.
        rng: np.random.RandomState
            random number generate to use for all randomness
        parallelize: bool
            Whether or not to execute each simulation in parallel.
        show_progress: bool
            Whether or not to display a progress-bar

    Returns
    -------
        observed_outcomes: np.ndarray
            Array of shape [num_agents] showing the observed outcome after
            performing the RCT experiment for all agents simultaneously.
        true_effects: np.ndarray
            Array of shape [num_agents] showing the `true` outcome for each
            agent. Here, true outcome means the outcome obtained by contrasting
            the experiment where no agent is treated and only agent i is treated.

    """
    baseline_assignments = np.zeros_like(treatment_assignments)
    counterfactual_assignments = []
    for idx in range(len(agents)):
        one_hot = np.zeros_like(treatment_assignments)
        one_hot[idx] = 1.0
        counterfactual_assignments.append(one_hot)

    assignments = [
        treatment_assignments,
        baseline_assignments,
    ] + counterfactual_assignments

    # Use the same seed for all runs to ensure difference in counterfactual
    # outcome only due to treatment assignment.
    seed = rng.randint(0, 99999)
    parallel_args = [(agents, config, assign, seed) for assign in assignments]

    if parallelize:
        runs = utils.parallelize(
            run_simulator, parallel_args, show_progress=show_progress
        )
    else:
        runs = [run_simulator(*args) for args in parallel_args]

    # RCT Outcomes
    observed_outcomes = runs[0].days_active.values

    # Baseline outcomes
    baseline = runs[1].days_active.values

    true_effects = []
    for idx, run in enumerate(runs[2:]):
        effect = run.days_active.values - baseline
        true_effects.append(effect[idx])

    true_effects = np.array(true_effects)

    return observed_outcomes, true_effects


def sample_agents(rng, num_samples, citizen_vision):
    """Generate covariates for each agent in the model."""
    agents = []
    for _ in range(num_samples):
        agent = civil_violence.Agent()
        agent.vision = citizen_vision
        agent.hardship = rng.uniform()
        agent.risk_aversion = 0.1
        agent.active_threshold = 0.1
        agent.legitimacy = rng.uniform()
        agents.append(agent)
    return agents


@parameter(
    name="citizen_vision",
    default=5,
    values=list(range(1, 7)),
    description="How many other citizens each agent can see.",
)
@parameter(
    name="agent_density",
    default=0.8,
    values=np.arange(0.05, 0.95, 0.1),
    description="How densely packed are agents on the grid.",
)
@parameter(
    name="cop_fraction",
    default=0.05,
    values=[0.1, 0.2, 0.3],
    description="number of cops = floor(num_samples * cop_fraction)",
)
@parameter(
    name="arrest_prob_constant",
    default=0.9,
    values=[0.6, 1.2, 2.3, 4.6],
    description="How strong the interaction between agents is.",
)
@parameter(
    name="prison_interaction",
    default=0.4,
    values=[0.01, 0.05, 0.1, 0.15, 0.2],
    description="Degree to which regime legitimacy is drawn toward the min while in prison",
)
def run_civil_violence(
    citizen_vision,
    agent_density,
    cop_fraction,
    arrest_prob_constant,
    prison_interaction,
    num_samples=100,
    seed=None,
    show_progress=False,
    parallelize=True,
):
    """Run an RCT experiment on ABM civil violence model.

    Each unit in the experiment is an agent in the model. Treatment corresponds to
    increasing the 'regime_legitimacy' belief of an agent. Although treatment is
    assigned randomly to each agent, the units interact, which complicates
    estimation of treatment effects. The `prison_interaction` term controls the
    strength of agent interaction.

    Outcomes count the number of days each agent spent in active state.

    Parameters
    ----------
        citizen_vision: int
            How many adjacent cells each agent can see.
        agent_density: float
            Density of agents on fixed grid size.
        cop_fraction: float
            1 / cop_fraction is number of agents for each cop on the grid.
        arrest_prob_constant: float
            How strong the effect of other agents is on perception of arrest probability.
        prison_interaction: float
            Degree to which agents share regime legitimacy beliefs in prison.
        num_samples: int
            Number of agents to use in the experiments.
        seed: int
            (Optional) Seed for all randomness in the experiment.
        show_progress: bool
            (Optional) Whether or not to display a progress-bar.
        parallelize: bool
            (Optional) Whether or not to execute each simulation in parallel.

    Returns
    -------
        covariates: np.ndarray
            Array of shape [num_agents, num_agent_covariates] of agent covariates
        treatment: np.ndarray
            Array of shape [num_agents] showing treatment assignment for each agent.
        outcome: np.ndarray
            Array of shape [num_agents] with observed outcome for each agent.
        true_effect: np.ndarray
            Array of shape [num_agents] with true (marginal) treatment effects for each agent.

    """
    rng = np.random.RandomState(seed)

    # Determine grid size
    num_cops = int(np.floor(cop_fraction * num_samples))
    num_cells = (num_samples + num_cops) / agent_density
    side_length = int(np.ceil(np.sqrt(num_cells)))

    simulator_config = civil_violence.Config(
        grid_height=side_length,
        grid_width=side_length,
        cop_fraction=cop_fraction,
        arrest_prob_constant=arrest_prob_constant,
        prison_interaction=prison_interaction,
    )

    # Generate population of agents
    agents = sample_agents(rng, num_samples, citizen_vision)

    # Assign treatment
    rct_assignments = rng.uniform(size=(num_samples,)) < 0.5

    outcomes, true_effects = run_abm_experiment(
        agents,
        rct_assignments,
        simulator_config,
        rng,
        parallelize=parallelize,
        show_progress=show_progress,
    )

    covariates = pd.DataFrame(dataclasses.asdict(a) for a in agents).values
    treatment = rct_assignments.astype(np.int64)

    return (covariates, treatment, outcomes), true_effects


#: RCT experiment that explores how violations of SUTVA impact inference.
RCT = GenericExperiment(
    name="civil_violence_rct",
    description="RCT experiment showing the effect of SUTVA violations on inference.",
    run_method=run_civil_violence,
)



"""Simulate and analyze runs from the Schelling segregation model."""
import dataclasses

from mesa.batchrunner import BatchRunner
import numpy as np

from whynot.simulators.schelling.model import Schelling


@dataclasses.dataclass
class Config:
    # pylint:disable-msg=too-few-public-methods
    """Parameterization of the Schelling dynamics."""

    #: Vertical size of the grid
    height: int = 10
    #: Horizontal size of the grid
    width: int = 10
    #: Agent density
    density: float = 0.8
    #: Fraction of "minority" agents
    minority_pc: float = 0.2
    #: Agent propensity to live near other agents of the same type.
    homophily: float = 5.0
    #: How much education changes homophily
    education_boost: float = -1.0
    #: What percentage of agent receive education
    education_pc: float = 0.0


def get_segregation(model):
    """Find the fraction of agents that only have neighbors of their same type."""
    segregated_agents = 0
    for agent in model.schedule.agents:
        segregated = True
        for neighbor in model.grid.neighbor_iter(agent.pos):
            if neighbor.type != agent.type:
                segregated = False
                break
        if segregated:
            segregated_agents += 1

    if model.schedule.get_agent_count() == 0:
        return 0.0

    return segregated_agents / model.schedule.get_agent_count()


def simulate(config, rollouts=10, seed=None):
    """Simulate repeated runs of the Schelling model for config.

    Parameters
    ----------
        config: whynot.simulators.Schelling
            Configuration of the grid, agent properties, and model dynamics.
        rollouts: int
            How many times to run the model for the same configuration.
        seed: int
            (Optional) Seed all randomness in rollouts

    Returns
    -------
        segregated_fraction: pd.Series
            What fraction of the agents are segrated at the end of the run
            for each rollout.

    """
    model_reporters = {"Segregated_Agents": get_segregation}

    rng = np.random.RandomState(seed)

    param_sweep = BatchRunner(
        Schelling,
        fixed_parameters=dataclasses.asdict(config),
        max_steps=200,
        model_reporters=model_reporters,
        display_progress=False,
        # Use a different seed for each rollout
        variable_parameters={"seed": rng.randint(9999999, size=rollouts)},
        # Single rollout for each seed
        iterations=1,
    )
    param_sweep.run_all()
    dataframe = param_sweep.get_model_vars_dataframe()
    # Currently just reports the fraction segregated.
    return dataframe.Segregated_Agents.mean()


if __name__ == "__main__":
    print(simulate(Config()))



"""Experiments for HIV simulator."""

from whynot.dynamics import DynamicsExperiment
from whynot.framework import parameter
from whynot.simulators import hiv


__all__ = ["get_experiments", "HIVRCT", "HIVConfounding"]


def get_experiments():
    """Return all experiments for HIV."""
    return [HIVRCT, HIVConfounding]


def sample_initial_states(rng):
    """Sample initial state by randomly perturbing the default state."""
    state = hiv.State()
    state.uninfected_T1 *= rng.uniform(low=0.95, high=1.05)
    state.infected_T1 *= rng.uniform(low=0.95, high=1.05)
    state.uninfected_T2 *= rng.uniform(low=0.95, high=1.05)
    state.infected_T2 *= rng.uniform(low=0.95, high=1.05)
    state.free_virus *= rng.uniform(low=0.95, high=1.05)
    state.immune_response *= rng.uniform(low=0.95, high=1.05)
    return state


##################
# RCT Experiment
##################
# pylint: disable-msg=invalid-name
#: Experiment on effect of increased drug efficacy on infected macrophages (cells/ml)
HIVRCT = DynamicsExperiment(
    name="HIVRCT",
    description="Study effect of increased drug efficacy on infected macrophages (cells/ml).",
    simulator=hiv,
    simulator_config=hiv.Config(epsilon_1=0.1, start_time=0, end_time=150),
    intervention=hiv.Intervention(time=100, epsilon_1=0.5),
    state_sampler=sample_initial_states,
    propensity_scorer=0.5,
    outcome_extractor=lambda run: run[149].infected_T2,
    covariate_builder=lambda run: run.initial_state.values(),
)


##########################
# Confounding Experiments
##########################


@parameter(
    name="treatment_bias",
    default=0.9,
    description="Treatment probability bias between more infected and less infected units.",
)
def hiv_confounded_propensity(untreated_run, treatment_bias):
    """Probability of treating each unit.

    We are more likely to treat units with high immune response and free virus.
    """
    if (
        untreated_run.initial_state.immune_response > 10
        and untreated_run.initial_state.free_virus > 1
    ):
        return treatment_bias

    return 1.0 - treatment_bias


# pylint: disable-msg=invalid-name
#: Experiment on effect of increased drug efficacy on infected macrophages with confounding
HIVConfounding = DynamicsExperiment(
    name="HIVConfounding",
    description=(
        "Study effect of increased drug efficacy on infected macrophages (cells/ml). "
        "Units with high immune response and free virus are more likely to be treated."
    ),
    simulator=hiv,
    simulator_config=hiv.Config(epsilon_1=0.1, start_time=0, end_time=150),
    intervention=hiv.Intervention(time=100, epsilon_1=0.5),
    state_sampler=sample_initial_states,
    propensity_scorer=hiv_confounded_propensity,
    outcome_extractor=lambda run: run[149].infected_T2,
    covariate_builder=lambda intervention, run: run.initial_state.values(),
)



"""Basic set of experiments on the LaLonde dataset."""
import os

import numpy as np
import pandas as pd

from whynot.framework import GenericExperiment, parameter

__all__ = ["get_experiments", "RandomResponse"]


##################
# Helper functions
##################


def load_dataset():
    """Load the LaLonde dataset."""
    dir_path = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(dir_path, "lalonde.csv")

    lalonde = pd.read_csv(data_path, index_col=0)

    # Remove outcome
    lalonde = lalonde.drop("re78", axis=1)

    return lalonde.rename(columns={"treat": "treatment"})


########################
# Experiment definitions
########################
def get_experiments():
    """Return all of the LaLonde experiments."""
    return [RandomResponse]


@parameter(
    name="hidden_dim",
    default=32,
    values=[8, 16, 32, 64, 128, 256, 512],
    description="hidden dimension of 2-layer ReLu network response.",
)
@parameter(
    name="alpha_scale",
    default=0.01,
    values=np.linspace(1e-4, 10, 10),
    description="Scale of the hidden-layer weights.",
)
def run_lalonde(
    num_samples,
    hidden_dim,
    alpha_scale,
    seed=None,
    parallelize=True,
    show_progress=False,
):
    # pylint:disable-msg=unused-argument
    """Generate data from the LaLonde dataset with a random response function.

    The covariates and treatment are both specified by the dataset, and the
    response function is a random 2-layer neural network with ReLu.

    Parameters
    ----------
        num_samples: int
            This parameter is ignored since the LaLonde dataset size is fixed.
        hidden_dim: int
            Hidden dimension of the relu network.
        alpha_scale: float
            Standard deviation of the final layer weights.
        seed: int
            Random seed used for all internal randomness
        parallelize: bool
            Ignored, but included for consistency with GenericExperiment API.
        show_progress: False
            Ignored, but included for consistency with GenericExperiment API.

    """
    rng = np.random.RandomState(seed)

    dataset = load_dataset()
    treatment = dataset.treatment.values.astype(np.int64)
    covariates = dataset.drop("treatment", axis=1).values

    # Define the networks
    num_inputs = covariates.shape[1]
    control_config = {
        "W": 0.05 * rng.randn(num_inputs, hidden_dim),
        "alpha": alpha_scale * rng.randn(hidden_dim, 1),
    }

    treatment_config = {
        "W": 0.05 * rng.randn(num_inputs, hidden_dim),
        "alpha": alpha_scale * rng.randn(hidden_dim, 1),
    }

    def get_effect(features, treatment):
        if treatment:
            config = treatment_config
        else:
            config = control_config
        return np.maximum(features.dot(config["W"]), 0).dot(config["alpha"])[:, 0]

    control_outcomes = get_effect(covariates, treatment=False)
    treatment_outcomes = get_effect(covariates, treatment=True)

    outcomes = np.copy(control_outcomes)
    treatment_idxs = np.where(treatment == 1.0)
    outcomes[treatment_idxs] = treatment_outcomes[treatment_idxs]

    return (covariates, treatment, outcomes), treatment_outcomes - control_outcomes


# pylint: disable-msg=invalid-name
#: Experiment simulating an outcome function on top of fixed LaLonde covariates.
RandomResponse = GenericExperiment(
    name="lalonde",
    description=(
        "An experiment on the LaLone dataset with fixed covariates "
        "and random 2-layer Relu NN for the response."
    ),
    run_method=run_lalonde,
)



"""Average treatment effect estimation with propensity score matching."""
from time import perf_counter

import numpy as np
import pandas as pd
import statsmodels.api as sm
from sklearn.neighbors import NearestNeighbors

from whynot.framework import InferenceResult


def estimate_treatment_effect(covariates, treatment, outcome):
    """Estimate treatment effects using propensity score matching.

    Parameters
    ----------
        covariates: `np.ndarray`
            Array of shape [num_samples, num_features] of features
        treatment:  `np.ndarray`
            Binary array of shape [num_samples]  indicating treatment status for each
            sample.
        outcome:  `np.ndarray`
            Array of shape [num_samples] containing the observed outcome for each sample.

    Returns
    -------
        result: `whynot.framework.InferenceResult`
            InferenceResult object for this procedure

    """
    start_time = perf_counter()

    # Compute propensity scores with logistic regression model.
    features = sm.add_constant(covariates, has_constant="add")
    logit = sm.Logit(treatment, features)
    model = logit.fit(disp=0)
    propensity_scores = model.predict(features)

    matched_treatment, matched_outcome, matched_weights = get_matched_dataset(
        treatment, propensity_scores, outcome
    )

    ate = compute_ate(matched_outcome, matched_treatment, matched_weights)

    # Bootstrap confidence intervals
    samples = []
    num_units = len(matched_treatment)
    for _ in range(1000):
        sample_idxs = np.random.choice(num_units, size=num_units, replace=True)
        samples.append(
            compute_ate(
                matched_outcome[sample_idxs],
                matched_treatment[sample_idxs],
                matched_weights[sample_idxs],
            )
        )
    conf_int = (np.quantile(samples, 0.025), np.quantile(samples, 0.975))
    stop_time = perf_counter()

    return InferenceResult(
        ate=ate,
        stderr=None,
        ci=conf_int,
        individual_effects=None,
        elapsed_time=stop_time - start_time,
    )


def get_matched_dataset(treatment, scores, outcome, num_neighbors=5):
    """Construct matched dataset for both control and treatment based on the provided score.

    Parameters
    ----------
        treatment:  `np.ndarray`
            Binary array of shape [num_samples]  indicating treatment status for each
            sample.
        scores:  `np.ndarray`
            Array of shape [num_samples] containing the score to match on for
            each sample, e.g. propensity scores.
        outcome:  `np.ndarray`
            Array of shape [num_samples] containing the observed outcome for each sample.
        num_neighbors: int
            (Optional) Number of units to match with.

    Returns
    -------
        matched_treatment:  `np.ndarray`
            Binary array indicating treatment status for each sample in the
            matched dataset.
        matched_outcome:  `np.ndarray`
            Array containing the observed outcome for each sample in the matched
            dataset.
        weights:  `np.ndarray`
            Array containing the weight for each sample in the matched dataset.
            The original units all have weight 1, and the matched units have
            uniform weight 1 / num_neighbors.

    """
    # Split into treatment and control groups
    num_samples = len(outcome)
    treated_scores = scores[treatment == 1.0].reshape(-1, 1)
    treated_outcomes = outcome[treatment == 1.0]
    control_scores = scores[treatment == 0.0].reshape(-1, 1)
    control_outcomes = outcome[treatment == 0.0]

    def match(base_scores, target_scores):
        """Match units in base to target and return the matched indices."""
        search = NearestNeighbors(metric="euclidean", n_neighbors=num_neighbors)
        search.fit(base_scores)
        matched = search.kneighbors(target_scores, return_distance=False)
        return matched.ravel()

    # Match control units to treatment and treatment units to controls.
    matched_control = match(control_scores, treated_scores)
    matched_treated = match(treated_scores, control_scores)

    # Construct an augmented dataset with the matched units
    matched_outcome = np.concatenate(
        [outcome, treated_outcomes[matched_treated], control_outcomes[matched_control]]
    )
    matched_treatment = np.concatenate(
        [treatment, np.ones_like(matched_treated), np.zeros_like(matched_control)]
    )
    weights = np.ones_like(matched_outcome)
    weights[num_samples:] = 1.0 / num_neighbors

    return matched_treatment, matched_outcome, weights


def compute_ate(outcome, treatment, weights):
    """Compute the (weighted) ATE."""
    treated_weights = weights[treatment == 1.0] / np.sum(weights[treatment == 1.0])
    control_weights = weights[treatment == 0.0] / np.sum(weights[treatment == 0.0])

    treated_outcome = np.sum(outcome[treatment == 1.0] * treated_weights)
    control_outcome = np.sum(outcome[treatment == 0.0] * control_weights)

    return treated_outcome - control_outcome



"""Implementation of Momoh and Fügenschuh environment for Zika prevention.

Momoh, Abdulfatai A., and Armin Fügenschuh. "Optimal control of intervention
strategies and cost effectiveness analysis for a Zika virus model." Operations
Research for Health Care 18 (2018): 99-111.

https://www.sciencedirect.com/science/article/pii/S2211692316301084#!
"""
import numpy as np

from whynot.gym import spaces
from whynot.gym.envs import ODEEnvBuilder, register
from whynot.simulators.zika import Config, Intervention, simulate, State


def get_intervention(action, time):
    """Return the intervention in the simulator required to take action."""
    treated_bednet_use, condom_use, treatment_of_infected, indoor_spray_use = action
    return Intervention(
        time=time,
        treated_bednet_use=treated_bednet_use,
        condom_use=condom_use,
        treatment_of_infected=treatment_of_infected,
        indoor_spray_use=indoor_spray_use,
    )


def get_reward(intervention, state, time):
    """Compute the reward based on the observed state and choosen intervention."""
    A_1, A_2, A_3 = 60, 500, 60
    C_1, C_2, C_3, C_4 = 25, 20, 30, 40
    discount = 4.0 / 365

    cost = (
        A_1 * state.asymptomatic_humans
        + A_2 * state.symptomatic_humans
        + A_3 * state.mosquito_population
    )

    cost += 0.5 * (
        C_1 * intervention.updates["treated_bednet_use"] ** 2
        + C_2 * intervention.updates["condom_use"] ** 2
        + C_3 * intervention.updates["treatment_of_infected"] ** 2
        + C_4 * intervention.updates["indoor_spray_use"] ** 2
    )
    return -cost * np.exp(-discount * time)


def observation_space():
    """Return observation space, the positive orthant."""
    state_dim = State.num_variables()
    state_space_low = np.zeros(state_dim)
    state_space_high = np.inf * np.ones(state_dim)
    return spaces.Box(state_space_low, state_space_high)


def action_space():
    """Return action space.

    There are four control variables in the model:
        - Treated bednet use
        - Condom use
        - Direct treatment of infected humans
        - Indoor residual spray use.

    """
    return spaces.Box(np.zeros(4), np.ones(4))


ZikaEnv = ODEEnvBuilder(
    simulate_fn=simulate,
    config=Config(),
    initial_state=State(),
    action_space=action_space(),
    observation_space=observation_space(),
    timestep=1.0,
    intervention_fn=get_intervention,
    reward_fn=get_reward,
)

register(
    id="Zika-v0", entry_point=ZikaEnv, max_episode_steps=200, reward_threshold=1e10,
)



from setuptools import setup, find_packages

setup(
    name="whynot",
    version="0.11.1",
    author="John Miller",
    author_email="miller_john@berkeley.edu",
    description="A framework for benchmarking causal inference.",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "autograd",
        "dataclasses; python_version<'3.7'",
        "gym",
        "mesa",
        "numpy",
        "networkx",
        "pandas",
        "pyomo",
        "py_mini_racer",
        "scipy",
        "sklearn",
        "statsmodels",
        "tqdm",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
    ],
    python_requires=">=3.6",
)



"""Tests for causal inference methods."""
import numpy as np
import pytest

import whynot as wn
from whynot.algorithms import ols, propensity_score_matching, propensity_weighted_ols


def generate_dataset(num_samples, num_features, true_ate=1.0, seed=1234):
    """Generate an observational dataset."""
    np.random.seed(seed)

    features = [0.25 * np.random.randn(num_samples, 1) for _ in range(num_features)]
    covariates = np.concatenate(features, axis=1)

    # Logisitic treatment probability
    arg = np.sum(covariates, axis=1) + np.random.randn(num_samples)
    prob = np.exp(arg) / (1.0 + np.exp(arg))
    treatment = np.random.binomial(1, prob)

    # Outcome is confounded by treatment
    # ATE is true_ate since the covariates are zero-mean.
    outcome = (np.sum(covariates, axis=1) + true_ate) * treatment
    outcome += np.random.randn(num_samples)

    return covariates, treatment.ravel(), outcome.ravel()


@pytest.mark.parametrize(
    "estimator",
    [ols, propensity_score_matching, propensity_weighted_ols],
    ids=["ols", "propensity_score_matching", "propensity_weighted_ols"],
)
def test_estimator(estimator):
    """Verify the estimator correctly computes treatment effects."""
    num_samples = 5000
    num_features = 3
    true_ate = 8

    covariates, treatment, outcome = generate_dataset(
        num_samples=num_samples, num_features=num_features, true_ate=true_ate
    )

    result = propensity_score_matching.estimate_treatment_effect(
        covariates, treatment, outcome
    )
    assert result.ci[0] <= result.ate <= result.ci[1]
    assert result.ci[0] <= true_ate <= result.ci[1]
    assert true_ate - 0.1 <= result.ate <= true_ate + 0.1


def test_causal_suite():
    """Integration test for the causal suite."""
    num_samples = 200
    num_features = 15

    covariates, treatment, outcomes = generate_dataset(
        num_samples=num_samples, num_features=num_features
    )

    result = wn.causal_suite(covariates, treatment, outcomes)
    assert "ols" in result
    assert "propensity_score_matching" in result
    assert "propensity_weighted_ols" in result



"""Environment builder for simulators based on dynamical systems."""
import inspect

from whynot.gym import Env
from whynot.gym.utils import seeding


class ODEEnvBuilder(Env):
    """Environment builder for simulators derived from dynamical systems."""

    def __init__(
        self,
        simulate_fn,
        config,
        action_space,
        observation_space,
        initial_state,
        intervention_fn,
        reward_fn,
        observation_fn=None,
        timestep=1.0,
    ):
        """Initialize an environment class.

        Parameters
        ----------
            simulate_fn:  Callable
                A function with signature
                simulate(initial_state, config, intervention=None, seed=None)
                -> whynot.dynamics.Run
            config: whynot.dynamics.BaseConfig
                The base simulator configuration
            action_space: whynot.gym.spaces.Space
                The action space for the reinforcement learner
            observation_space: whynot.gym.spaces.Space
                The space of observations for the agent
            initial_state: whynot.dynamics.BaseState
                The initial state of the simulator
            intervention_fn: Callable
                A function that maps actions to simulator interventions with signature
                get_intervention(action, time) -> whynot.dynamics.BaseState
            reward_fn: Callable
                A function that computes the cost/reward of taking
                an intervention in a particular state state with signature
                get_reward(intervention, state) -> float
            observation_fn: Callable
                (Optional) A function that computes the observed state for the
                state of the simulator with signature
                observation_fn(state) -> np.ndarray.
                If ommitted, the entire simulator state is returned.
            timestep: float
                Time between successive observations in the dynamical system.

        """
        self.config = config
        self.action_space = action_space
        self.observation_space = observation_space

        self.initial_state = initial_state
        self.state = self.initial_state

        self.simulate_fn = simulate_fn

        self.start_time = self.config.start_time
        self.terminal_time = self.config.end_time
        self.timestep = timestep
        self.time = self.start_time

        self.intervention_fn = intervention_fn
        self.reward_fn = reward_fn

        self.seed()

    def reset(self):
        """Reset the state."""
        self.state = self.initial_state
        self.time = self.start_time
        return self._get_observation(self.state)

    def seed(self, seed=None):
        """Set internal randomness of the environment."""
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action):
        """Perform one forward step in the environment.

        Parameters
        ----------
            action: A numpy array reprsenting an action of shape
                [1, action_dim].

        Returns
        -------
            observation: A numpy array of shape [1, obs_dim].
            reward: A numpy array of shape [1, 1].
            done: A numpy array of shape [1, 1]
            info_dict: An empty dict.

        """
        if not self.action_space.contains(action):
            raise ValueError("%r (%s) invalid" % (action, type(action)))
        intervention = self.intervention_fn(action, self.time)
        # Set the start and end time in config to simulate one timestep.
        self.config.start_time = self.time
        self.config.end_time = self.time + self.timestep
        self.time += self.timestep
        # Get the next state from simulation.
        self.state = self.simulate_fn(
            initial_state=self.state, config=self.config, intervention=intervention
        )[self.time]
        done = bool(self.time >= self.terminal_time)
        reward = self._get_reward(intervention, self.state)
        return self._get_observation(self.state), reward, done, {}

    def render(self, mode="human"):
        """Render the environment, unused."""

    @staticmethod
    def _get_observation(state):
        """Convert a state to a numpy array observation.

        By default, returns the fully observed state in order listed in the
        State class.

        Parameters
        ----------
            state: An instance of whynot.dynamics.BaseState.

        Returns
        -------
            A numpy array of shape [1, obs_dim].

        """
        return state.values()

    @staticmethod
    def _get_args(func):
        """Return the arguments to the function."""
        return inspect.signature(func).parameters

    def _get_reward(self, intervention, state):
        """Return the reward obtained by intervening in the given state."""
        reward_args = self._get_args(self.reward_fn)
        kwargs = {}
        if "config" in reward_args:
            kwargs["config"] = self.config
        if "time" in reward_args:
            kwargs["time"] = self.time
        return self.reward_fn(intervention=intervention, state=state, **kwargs)

    def __call__(self):
        """Return the class, as if this function were calling the constructor."""
        self.reset()
        return self



"""Unit tests for world 3 model."""
from whynot.simulators.world3.simulator import *


def get_world3_context():
    """Load a javascript context for world3"""
    ctx = PyMiniRacerContext()
    ctx.eval(WORLD3_JS_CODE)
    return ctx


def test_set_state():
    """Check setting initial_state."""
    ctx = get_world3_context()
    initial_state = State(land_fertility=3434)

    set_state(ctx, initial_state)
    assert ctx.eval(f"landFertility.initVal") == 3434

    # Run the model and check if land fertility is correctly used.
    ctx.eval("fastRun()")
    states, _ = decode_states(ctx)
    assert states[0].land_fertility == 3434


def test_set_config():
    """Check setting config + intervention works as expected."""
    ctx = get_world3_context()
    config = Config(service_capital_output_ratio=23)
    intervention = Intervention(nonrenewable_resource_usage_factor=2)

    set_config(ctx, config, intervention)

    for param in dataclasses.asdict(config):
        if param in ["policy_year", "start_time", "end_time", "delta_t"]:
            continue

        if param == "service_capital_output_ratio":
            assert ctx.eval("serviceCapitalOutputRatio.before") == 23
            assert ctx.eval("serviceCapitalOutputRatio.after") == 23
        elif param == "nonrenewable_resource_usage_factor":
            assert (
                ctx.eval("nonrenewableResourceUsageFactor.before")
                == config.nonrenewable_resource_usage_factor
            )
            assert ctx.eval("nonrenewableResourceUsageFactor.after") == 2
        else:
            assert ctx.eval(f"{to_camel_case(param)}.before") == ctx.eval(
                f"{to_camel_case(param)}.after"
            )


def test_setup():
    """Test setup of world3 simulator."""
    for idx in range(3):
        ctx = get_world3_context()
        initial_state = State(land_fertility=idx * 1111)
        config = Config(service_capital_output_ratio=idx * 22)
        intervention = Intervention(nonrenewable_resource_usage_factor=idx * 22)

        set_state(ctx, initial_state)
        set_config(ctx, config, intervention)

        assert ctx.eval(f"landFertility.initVal") == idx * 1111
        assert ctx.eval("serviceCapitalOutputRatio.before") == idx * 22
        assert ctx.eval("serviceCapitalOutputRatio.after") == idx * 22
        assert (
            ctx.eval("nonrenewableResourceUsageFactor.before")
            == config.nonrenewable_resource_usage_factor
        )
        assert ctx.eval("nonrenewableResourceUsageFactor.after") == idx * 22
